# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import os
import time

from tacker.objects import fields
from tacker.tests.functional.sol_v2 import base_v2
from tacker.tests.functional.sol_v2 import paramgen


@ddt.ddt
class VnfLcmErrorHandlingTest(base_v2.BaseSolV2Test):

    @classmethod
    def setUpClass(cls):
        super(VnfLcmErrorHandlingTest, cls).setUpClass()

        cur_dir = os.path.dirname(__file__)
        # tacker/tests/etc...
        #             /functional/sol_v2
        image_dir = os.path.join(
            cur_dir, "../../etc/samples/etsi/nfv/common/Files/images")
        image_file = "cirros-0.5.2-x86_64-disk.img"
        image_path = os.path.abspath(os.path.join(image_dir, image_file))

        # Scale operation will fail
        scale_ng_path = os.path.join(cur_dir, "samples/scale_ng")
        cls.vnf_pkg_1, cls.vnfd_id_1 = cls.create_vnf_package(
            scale_ng_path, image_path=image_path)

        # Instantiate VNF will fail
        error_network_path = os.path.join(cur_dir, "samples/error_network")
        # no image contained
        cls.vnf_pkg_2, cls.vnfd_id_2 = cls.create_vnf_package(
            error_network_path)

        # update VNF or change external VNF connectivity will fail
        update_change_ng_path = os.path.join(cur_dir,
                                             "samples/basic_lcms_min")
        # no image contained
        cls.vnf_pkg_3, cls.vnfd_id_3 = cls.create_vnf_package(
            update_change_ng_path)

    @classmethod
    def tearDownClass(cls):
        super(VnfLcmErrorHandlingTest, cls).tearDownClass()

        cls.delete_vnf_package(cls.vnf_pkg_1)
        cls.delete_vnf_package(cls.vnf_pkg_2)
        cls.delete_vnf_package(cls.vnf_pkg_3)

    def setUp(self):
        super().setUp()

    def test_retry_rollback_scale_out(self):
        """Test retry and rollback scale out operations

        * About attributes:
          All of the following cardinality attributes are set.
          In addition, 0..N or 1..N attributes are set to 2 or more.
          0..1 is set to 1.
          - 0..1 (1)
          - 0..N (2 or more)
          - 1..N (2 or more)

        * About LCM operations:
          This test includes the following operations.
          - 0. Pre-setting
          - 1. Create subscription
          - 2. Test notification
          - 3. Create VNF instance
          - 4. Instantiate VNF
          - 5. Show VNF instance
          - 6. Scale out operation(will fail)
          - 7. Show VNF instance
          - 8. Retry operation
          - 9. Rollback scale out operation
          - 10. Show VNF LCM operation occurrence
          - 11. List VNF LCM operation occurrence
          - 12. Terminate VNF
          - 13. Delete VNF instance
          - 14. Delete subscription
          - 15. Show subscription
        """
        # 0. Pre-setting
        # Create a new network and subnet to check the IP allocation of
        # IPv4 and IPv6
        ft_net0_name = 'ft-net0'
        ft_net0_subs = {
            'ft-ipv4-subnet0': {
                'range': '100.100.100.0/24',
                'ip_version': 4
            },
            'ft-ipv6-subnet0': {
                'range': '1111:2222:3333::/64',
                'ip_version': 6
            }
        }
        ft_net0_id = self.create_network(ft_net0_name)
        self.addCleanup(self.delete_network, ft_net0_id)
        for sub_name, val in ft_net0_subs.items():
            # subnet is automatically deleted with network deletion
            self.create_subnet(
                ft_net0_id, sub_name, val['range'], val['ip_version'])

        net_ids = self.get_network_ids(
            ['net0', 'net1', 'net_mgmt', 'ft-net0'])
        subnet_ids = self.get_subnet_ids(
            ['subnet0', 'subnet1', 'ft-ipv4-subnet0', 'ft-ipv6-subnet0'])

        port_names = ['VDU2_CP1-1', 'VDU2_CP1-2']
        port_ids = {}
        for port_name in port_names:
            port_id = self.create_port(net_ids['net0'], port_name)
            port_ids[port_name] = port_id
            self.addCleanup(self.delete_port, port_id)

        # 1. Create subscription
        callback_url = os.path.join(base_v2.MOCK_NOTIFY_CALLBACK_URL,
                                    self._testMethodName)
        callback_uri = ('http://localhost:'
                        f'{base_v2.FAKE_SERVER_MANAGER.SERVER_PORT}'
                        f'{callback_url}')

        sub_req = paramgen.sub_create_max(callback_uri)
        resp, body = self.create_subscription(sub_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        sub_id = body['id']

        # 2. Test notification
        self.assert_notification_get(callback_url)
        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 3. Create VNF instance
        # ETSI NFV SOL003 v3.3.1 5.5.2.2 VnfInstance
        expected_inst_attrs = [
            'id',
            'vnfInstanceName',
            'vnfInstanceDescription',
            'vnfdId',
            'vnfProvider',
            'vnfProductName',
            'vnfSoftwareVersion',
            'vnfdVersion',
            # 'vnfConfigurableProperties', # omitted
            # 'vimConnectionInfo', # omitted
            'instantiationState',
            # 'instantiatedVnfInfo', # omitted
            'metadata',
            # 'extensions', # omitted
            '_links'
        ]
        create_req = paramgen.create_vnf_max(self.vnfd_id_1)
        resp, body = self.create_vnf_instance(create_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        self.check_resp_body(body, expected_inst_attrs)
        inst_id = body['id']

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body['instantiationState'])

        # 4. Instantiate VNF
        instantiate_req = paramgen.instantiate_vnf_max(
            net_ids, subnet_ids, port_ids, self.auth_url)
        resp, body = self.instantiate_vnf_instance(inst_id, instantiate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 5. Show VNF instance
        additional_inst_attrs = [
            'vimConnectionInfo',
            'instantiatedVnfInfo',
            'extensions',
            'vnfConfigurableProperties'
        ]
        expected_inst_attrs.extend(additional_inst_attrs)
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.INSTANTIATED,
                         body.get('instantiationState'))

        # check vnfState of VNF
        self.assertEqual(fields.VnfOperationalStateType.STARTED,
                         body['instantiatedVnfInfo']['vnfState'])

        # 6. Scale out operation(will fail)
        scaleout_req = paramgen.scaleout_vnf_max()
        resp, body = self.scale_vnf_instance(inst_id, scaleout_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)
        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 7. Show VNF instance
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # 8. Retry scale out operation
        resp, body = self.retry_lcmocc(lcmocc_id)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_delete(resp)
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # 9. Rollback scale out operation
        resp, body = self.rollback_lcmocc(lcmocc_id)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_delete(resp)
        self.wait_lcmocc_rolled_back(lcmocc_id)

        # 10. Show VNF LCM operation occurrence
        # ETSI NFV SOL003 v3.3.1 5.5.2.13 VnfLcmOpOcc
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            'grantId',
            'operation',
            'isAutomaticInvocation',
            'operationParams',
            'isCancelPending',
            # 'cancelMode', # omitted
            'error',
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.show_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)

        # 11. List VNF LCM operation occurrence
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            # 'grantId', # omitted
            'operation',
            'isAutomaticInvocation',
            # 'operationParams', # omitted
            'isCancelPending',
            # 'cancelMode', # omitted
            # 'error', # omitted
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.list_lcmocc()
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        for lcmocc in body:
            self.check_resp_body(lcmocc, expected_attrs)

        # 12. Terminate VNF instance
        terminate_req = paramgen.terminate_vnf_max()
        resp, body = self.terminate_vnf_instance(inst_id, terminate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body.get('instantiationState'))

        # wait a bit because there is a bit time lag between vnf instance DB
        # terminate and delete completion.
        time.sleep(5)

        # 13. Delete VNF instance
        resp, body = self.delete_vnf_instance(inst_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_1)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 14. Delete subscription
        resp, body = self.delete_subscription(sub_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # 15. Show subscription
        resp, body = self.show_subscription(sub_id)
        self.assertEqual(404, resp.status_code)
        self.check_resp_headers_in_get(resp)

    def test_rollback_instantiate(self):
        """Test rollback instantiate operation

        * About attributes:
          Omit except for required attributes.
          Only the following cardinality attributes are set.
          - 1
          - 1..N (1)

        * About LCM operations:
          This test includes the following operations.
          - 1. Create subscription
          - 2. Test notification
          - 3. Create VNF instance
          - 4. Instantiate VNF(will fail)
          - 5. Show VNF instance
          - 6. Rollback instantiation operation
          - 7. Show VNF LCM operation occurrence
          - 8. List VNF LCM operation occurrence
          - 9. Delete VNF instance
          - 10. Delete subscription
        """

        # 1. Create subscription
        callback_url = os.path.join(base_v2.MOCK_NOTIFY_CALLBACK_URL,
                                    self._testMethodName)
        callback_uri = ('http://localhost:'
                        f'{base_v2.FAKE_SERVER_MANAGER.SERVER_PORT}'
                        f'{callback_url}')

        sub_req = paramgen.sub_create_min(callback_uri)
        resp, body = self.create_subscription(sub_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        sub_id = body['id']

        # 2. Test notification
        self.assert_notification_get(callback_url)
        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 3. Create VNF instance
        # ETSI NFV SOL003 v3.3.1 5.5.2.2 VnfInstance
        expected_inst_attrs = [
            'id',
            # 'vnfInstanceName', # omitted
            # 'vnfInstanceDescription', # omitted
            'vnfdId',
            'vnfProvider',
            'vnfProductName',
            'vnfSoftwareVersion',
            'vnfdVersion',
            # 'vnfConfigurableProperties', # omitted
            # 'vimConnectionInfo', # omitted
            'instantiationState',
            # 'instantiatedVnfInfo', # omitted
            # 'metadata', # omitted
            # 'extensions', # omitted
            '_links'
        ]
        create_req = paramgen.create_vnf_min(self.vnfd_id_2)
        resp, body = self.create_vnf_instance(create_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        self.check_resp_body(body, expected_inst_attrs)
        inst_id = body['id']

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body.get('instantiationState'))

        # 4. Instantiate VNF(will fail)
        instantiate_req = paramgen.instantiate_vnf_min()
        resp, body = self.instantiate_vnf_instance(inst_id, instantiate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 5. Show VNF instance
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body.get('instantiationState'))

        # 6. Rollback instantiation operation
        resp, body = self.rollback_lcmocc(lcmocc_id)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_delete(resp)
        self.wait_lcmocc_rolled_back(lcmocc_id)

        # 7. Show VNF LCM operation occurrence
        # ETSI NFV SOL003 v3.3.1 5.5.2.13 VnfLcmOpOcc
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            'grantId',
            'operation',
            'isAutomaticInvocation',
            'operationParams',
            'isCancelPending',
            # 'cancelMode', # omitted
            'error',
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.show_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)

        # 8. List VNF LCM operation occurrence
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            # 'grantId', # omitted
            'operation',
            'isAutomaticInvocation',
            # 'operationParams', # omitted
            'isCancelPending',
            # 'cancelMode', # omitted
            # 'error', # omitted
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.list_lcmocc()
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        for lcmocc in body:
            self.check_resp_body(lcmocc, expected_attrs)

        # 9. Delete VNF instance
        resp, body = self.delete_vnf_instance(inst_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 10. Delete subscription
        resp, body = self.delete_subscription(sub_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

    def test_fail_instantiate(self):
        """Test fail instantiate operation

        * About attributes:
          Omit except for required attributes.
          Only the following cardinality attributes are set.
          - 1
          - 1..N (1)

        * About LCM operations:
          This test includes the following operations.
          - 1. Create subscription
          - 2. Test notification
          - 3. Create VNF instance
          - 4. Instantiate VNF(will fail)
          - 5. Show VNF instance
          - 6. Fail instantiation operation
          - 7. Show VNF LCM operation occurrence
          - 8. List VNF LCM operation occurrence
          - 9. Delete VNF instance
          - 10. Delete subscription
        """

        # 1. Create subscription
        callback_url = os.path.join(base_v2.MOCK_NOTIFY_CALLBACK_URL,
                                    self._testMethodName)
        callback_uri = ('http://localhost:'
                        f'{base_v2.FAKE_SERVER_MANAGER.SERVER_PORT}'
                        f'{callback_url}')

        sub_req = paramgen.sub_create_min(callback_uri)
        resp, body = self.create_subscription(sub_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        sub_id = body['id']

        # 2. Test notification
        self.assert_notification_get(callback_url)
        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 3. Create VNF instance
        # ETSI NFV SOL003 v3.3.1 5.5.2.2 VnfInstance
        expected_inst_attrs = [
            'id',
            # 'vnfInstanceName', # omitted
            # 'vnfInstanceDescription', # omitted
            'vnfdId',
            'vnfProvider',
            'vnfProductName',
            'vnfSoftwareVersion',
            'vnfdVersion',
            # 'vnfConfigurableProperties', # omitted
            # 'vimConnectionInfo', # omitted
            'instantiationState',
            # 'instantiatedVnfInfo', # omitted
            # 'metadata', # omitted
            # 'extensions', # omitted
            '_links'
        ]
        create_req = paramgen.create_vnf_min(self.vnfd_id_2)
        resp, body = self.create_vnf_instance(create_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        self.check_resp_body(body, expected_inst_attrs)
        inst_id = body['id']

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body.get('instantiationState'))

        # 4. Instantiate VNF(will fail)
        instantiate_req = paramgen.instantiate_vnf_min()
        resp, body = self.instantiate_vnf_instance(inst_id, instantiate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 5. Show VNF instance
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body.get('instantiationState'))

        # 6. Fail instantiation operation
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            'grantId',
            'operation',
            'isAutomaticInvocation',
            'operationParams',
            'isCancelPending',
            # 'cancelMode', # omitted
            'error',
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.fail_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)
        self.assertEqual('FAILED', body['operationState'])

        # 7. Show VNF LCM operation occurrence
        # ETSI NFV SOL003 v3.3.1 5.5.2.13 VnfLcmOpOcc
        resp, body = self.show_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)

        # 8. List VNF LCM operation occurrence
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            # 'grantId', # omitted
            'operation',
            'isAutomaticInvocation',
            # 'operationParams', # omitted
            'isCancelPending',
            # 'cancelMode', # omitted
            # 'error', # omitted
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.list_lcmocc()
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        for lcmocc in body:
            self.check_resp_body(lcmocc, expected_attrs)

        # 9. Delete VNF instance
        # Delete Stack
        self.heat_client.delete_stack(f'vnf-{inst_id}')

        resp, body = self.delete_vnf_instance(inst_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_2)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 10. Delete subscription
        resp, body = self.delete_subscription(sub_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

    def test_rollback_update(self):
        """Test rollback update VNF operation

        * About attributes:
          Omit except for required attributes.
          Only the following cardinality attributes are set.
          - 1
          - 1..N (1)

        * About LCM operations:
          This test includes the following operations.
          - 1. Create subscription
          - 2. Test notification
          - 3. Create VNF instance
          - 4. Instantiate VNF
          - 5. Show VNF instance
          - 6. Update VNF(will fail)
          - 7. Rollback update operation
          - 8. Show VNF LCM operation occurrence
          - 9. List VNF LCM operation occurrence
          - 10. Terminate VNF
          - 11. Delete VNF instance
          - 12. Delete subscription
        """

        # 1. Create subscription
        callback_url = os.path.join(base_v2.MOCK_NOTIFY_CALLBACK_URL,
                                    self._testMethodName)
        callback_uri = ('http://localhost:'
                        f'{base_v2.FAKE_SERVER_MANAGER.SERVER_PORT}'
                        f'{callback_url}')

        sub_req = paramgen.sub_create_min(callback_uri)
        resp, body = self.create_subscription(sub_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        sub_id = body['id']

        # 2. Test notification
        self.assert_notification_get(callback_url)
        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 3. Create VNF instance
        # ETSI NFV SOL003 v3.3.1 5.5.2.2 VnfInstance
        expected_inst_attrs = [
            'id',
            # 'vnfInstanceName', # omitted
            # 'vnfInstanceDescription', # omitted
            'vnfdId',
            'vnfProvider',
            'vnfProductName',
            'vnfSoftwareVersion',
            'vnfdVersion',
            # 'vnfConfigurableProperties', # omitted
            # 'vimConnectionInfo', # omitted
            'instantiationState',
            # 'instantiatedVnfInfo', # omitted
            # 'metadata', # omitted
            # 'extensions', # omitted
            '_links'
        ]
        create_req = paramgen.create_vnf_min(self.vnfd_id_3)
        resp, body = self.create_vnf_instance(create_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        self.check_resp_body(body, expected_inst_attrs)
        inst_id = body['id']

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body['instantiationState'])

        # 4. Instantiate VNF
        instantiate_req = paramgen.instantiate_vnf_min()
        resp, body = self.instantiate_vnf_instance(inst_id, instantiate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 5. Show VNF instance
        additional_inst_attrs = [
            'vimConnectionInfo',
            'instantiatedVnfInfo'
        ]
        expected_inst_attrs.extend(additional_inst_attrs)
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.INSTANTIATED,
                         body['instantiationState'])

        # check vnfState of VNF
        self.assertEqual(fields.VnfOperationalStateType.STARTED,
                         body['instantiatedVnfInfo']['vnfState'])

        # 6. Update VNF(will fail)
        # NOTE: Create a file so that an error occurs in mgmtDriver
        path = '/tmp/modify_information_start'
        with open(path, 'w', encoding='utf-8') as f:
            f.write('')
        self.addCleanup(os.remove, path)
        update_req = paramgen.update_vnf_min()
        resp, body = self.update_vnf_instance(inst_id, update_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # 7. Rollback update operation
        resp, body = self.rollback_lcmocc(lcmocc_id)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_delete(resp)
        self.wait_lcmocc_rolled_back(lcmocc_id)

        # 8. Show VNF LCM operation occurrence
        # ETSI NFV SOL003 v3.3.1 5.5.2.13 VnfLcmOpOcc
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            # 'grantId', # omitted
            'operation',
            'isAutomaticInvocation',
            # 'operationParams', # omitted
            'isCancelPending',
            # 'cancelMode', # omitted
            # 'error', # omitted
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.show_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)

        # 9. List VNF LCM operation occurrence
        resp, body = self.list_lcmocc()
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        for lcmocc in body:
            self.check_resp_body(lcmocc, expected_attrs)

        # 10. Terminate a VNF instance
        terminate_req = paramgen.terminate_vnf_min()
        resp, body = self.terminate_vnf_instance(inst_id, terminate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # wait a bit because there is a bit time lag between lcmocc DB
        # update and terminate completion.
        time.sleep(10)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body['instantiationState'])

        # 11. Delete VNF instance
        resp, body = self.delete_vnf_instance(inst_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 12. Delete subscription
        resp, body = self.delete_subscription(sub_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

    def test_rollback_chgextconn(self):
        """Test rollback change_ext_conn operation

        * About attributes:
          All of the following cardinality attributes are set.
          In addition, 0..N or 1..N attributes are set to 2 or more.
          0..1 is set to 1.
          - 0..1 (1)
          - 0..N (2 or more)
          - 1..N (2 or more)

        * About LCM operations:
          This test includes the following operations.
          - 0. Pre-setting
          - 1. Create subscription
          - 2. Test notification
          - 3. Create VNF instance
          - 4. Instantiate VNF
          - 5. Show VNF instance
          - 6. Change external connectivity(will fail)
          - 7. Rollback change_ext_conn operation
          - 8. Show VNF LCM operation occurrence
          - 9. List VNF LCM operation occurrence
          - 10. Terminate VNF
          - 11. Delete VNF instance
          - 12. Delete subscription
        """
        # 0. Pre-setting
        # Create a new network for change external connectivity
        ft_net1_name = 'ft-net1'
        ft_net1_subs = {
            'ft-ipv4-subnet1': {
                'range': '22.22.22.0/24',
                'ip_version': 4
            },
            'ft-ipv6-subnet1': {
                'range': '1111:2222:4444::/64',
                'ip_version': 6
            }
        }
        ft_net1_id = self.create_network(ft_net1_name)
        self.addCleanup(self.delete_network, ft_net1_id)
        for sub_name, val in ft_net1_subs.items():
            # subnet is automatically deleted with network deletion
            self.create_subnet(
                ft_net1_id, sub_name, val['range'], val['ip_version'])

        net_ids = self.get_network_ids(['ft-net1'])
        subnet_ids = self.get_subnet_ids(['ft-ipv4-subnet1',
            'ft-ipv6-subnet1'])

        # 1. Create subscription
        callback_url = os.path.join(base_v2.MOCK_NOTIFY_CALLBACK_URL,
                                    self._testMethodName)
        callback_uri = ('http://localhost:'
                        f'{base_v2.FAKE_SERVER_MANAGER.SERVER_PORT}'
                        f'{callback_url}')

        sub_req = paramgen.sub_create_min(callback_uri)
        resp, body = self.create_subscription(sub_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        sub_id = body['id']

        # 2. Test notification
        self.assert_notification_get(callback_url)
        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 3. Create VNF instance
        # ETSI NFV SOL003 v3.3.1 5.5.2.2 VnfInstance
        expected_inst_attrs = [
            'id',
            # 'vnfInstanceName', # omitted
            # 'vnfInstanceDescription', # omitted
            'vnfdId',
            'vnfProvider',
            'vnfProductName',
            'vnfSoftwareVersion',
            'vnfdVersion',
            # 'vnfConfigurableProperties', # omitted
            # 'vimConnectionInfo', # omitted
            'instantiationState',
            # 'instantiatedVnfInfo', # omitted
            # 'metadata', # omitted
            # 'extensions', # omitted
            '_links'
        ]
        create_req = paramgen.create_vnf_min(self.vnfd_id_3)
        resp, body = self.create_vnf_instance(create_req)
        self.assertEqual(201, resp.status_code)
        self.check_resp_headers_in_create(resp)
        self.check_resp_body(body, expected_inst_attrs)
        inst_id = body['id']

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body['instantiationState'])

        # 4. Instantiate VNF
        instantiate_req = paramgen.instantiate_vnf_min()
        resp, body = self.instantiate_vnf_instance(inst_id, instantiate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # 5. Show VNF instance
        additional_inst_attrs = [
            'vimConnectionInfo',
            'instantiatedVnfInfo'
        ]
        expected_inst_attrs.extend(additional_inst_attrs)
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_inst_attrs)

        # check instantiationState of VNF
        self.assertEqual(fields.VnfInstanceState.INSTANTIATED,
                         body['instantiationState'])

        # check vnfState of VNF
        self.assertEqual(fields.VnfOperationalStateType.STARTED,
                         body['instantiatedVnfInfo']['vnfState'])

        # 6. Change external connectivity(will fail)
        # NOTE: Create a file so that an error occurs in mgmtDriver
        path = '/tmp/change_external_connectivity_start'
        with open(path, 'w', encoding='utf-8') as f:
            f.write('')
        self.addCleanup(os.remove, path)
        change_ext_conn_req = paramgen.change_ext_conn_min(net_ids, subnet_ids)
        resp, body = self.change_ext_conn(inst_id, change_ext_conn_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)
        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_failed_temp(lcmocc_id)

        # 7. Rollback change_ext_conn operation
        resp, body = self.rollback_lcmocc(lcmocc_id)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_delete(resp)
        self.wait_lcmocc_rolled_back(lcmocc_id)

        # 8. Show VNF LCM operation occurrence
        # ETSI NFV SOL003 v3.3.1 5.5.2.13 VnfLcmOpOcc
        # NOTE: omitted values are not supported at that time
        expected_attrs = [
            'id',
            'operationState',
            'stateEnteredTime',
            'startTime',
            'vnfInstanceId',
            # 'grantId', # omitted
            'operation',
            'isAutomaticInvocation',
            # 'operationParams', # omitted
            'isCancelPending',
            # 'cancelMode', # omitted
            # 'error', # omitted
            # 'resourceChanges', # omitted
            # 'changedInfo', # omitted
            # 'changedExtConnectivity', # omitted
            # 'modificationsTriggeredByVnfPkgChange', # omitted
            # 'vnfSnapshotInfoId', # omitted
            '_links'
        ]
        resp, body = self.show_lcmocc(lcmocc_id)
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        self.check_resp_body(body, expected_attrs)

        # 9. List VNF LCM operation occurrence
        resp, body = self.list_lcmocc()
        self.assertEqual(200, resp.status_code)
        self.check_resp_headers_in_get(resp)
        for lcmocc in body:
            self.check_resp_body(lcmocc, expected_attrs)

        # 10. Terminate VNF
        terminate_req = paramgen.terminate_vnf_min()
        resp, body = self.terminate_vnf_instance(inst_id, terminate_req)
        self.assertEqual(202, resp.status_code)
        self.check_resp_headers_in_operation_task(resp)

        lcmocc_id = os.path.basename(resp.headers['Location'])
        self.wait_lcmocc_complete(lcmocc_id)

        # wait a bit because there is a bit time lag between lcmocc DB
        # update and terminate completion.
        time.sleep(10)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('IN_USE', usage_state)

        # check instantiationState of VNF
        resp, body = self.show_vnf_instance(inst_id)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(fields.VnfInstanceState.NOT_INSTANTIATED,
                         body['instantiationState'])

        # 11. Delete VNF instance
        resp, body = self.delete_vnf_instance(inst_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)

        # check usageState of VNF Package
        usage_state = self.get_vnf_package(self.vnf_pkg_3)['usageState']
        self.assertEqual('NOT_IN_USE', usage_state)

        # 12. Delete subscription
        resp, body = self.delete_subscription(sub_id)
        self.assertEqual(204, resp.status_code)
        self.check_resp_headers_in_delete(resp)



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import timeutils

from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.extensions import common_services
from tacker.plugins.common_services import common_services_plugin
from tacker.tests.unit.db import base as db_base


class TestCommonServicesPlugin(db_base.SqlTestCase):
    def setUp(self):
        super(TestCommonServicesPlugin, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self.event_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        self.coreutil_plugin = common_services_plugin.CommonServicesPlugin()

    def _get_dummy_event_obj(self):
        return {
            'resource_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            'resource_state': 'ACTIVE',
            'resource_type': 'VNF',
            'event_details': '',
            'event_type': 'scale_up',
            'timestamp': timeutils.parse_strtime('2016-07-20T05:43:52.765172')
        }

    def test_create_event(self):
        evt_obj = self._get_dummy_event_obj()
        result = self.event_db_plugin.create_event(self.context,
                                                   evt_obj['resource_id'],
                                                   evt_obj['resource_type'],
                                                   evt_obj['resource_state'],
                                                   evt_obj['event_type'],
                                                   evt_obj['timestamp'],
                                                   evt_obj['event_details'])
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('resource_id', result)
        self.assertIn('resource_state', result)
        self.assertIn('resource_type', result)
        self.assertIn('event_type', result)
        self.assertIn('event_details', result)
        self.assertIn('timestamp', result)

    def test_event_not_found(self):
        self.assertRaises(common_services.EventNotFoundException,
                          self.coreutil_plugin.get_event, self.context, '99')

    def test_InvalidModelInputExceptionNotThrown(self):
        evt_obj = self._get_dummy_event_obj()
        result = self.event_db_plugin.create_event(self.context,
                                                   evt_obj['resource_id'],
                                                   evt_obj['resource_type'],
                                                   evt_obj['resource_state'],
                                                   evt_obj['event_type'],
                                                   evt_obj['timestamp'],
                                                   evt_obj['event_details'])
        try:
            self.coreutil_plugin.get_event(self.context, str(result['id']))
        except common_services.InvalidModelException:
            self.assertTrue(False)
        except Exception:
            self.assertTrue(True)

    def test_get_event_by_id(self):
        evt_obj = self._get_dummy_event_obj()
        evt_created = self.event_db_plugin.create_event(
            self.context, evt_obj['resource_id'],
            evt_obj['resource_type'],
            evt_obj['resource_state'],
            evt_obj['event_type'],
            evt_obj['timestamp'],
            evt_obj['event_details'])
        self.assertIsNotNone(evt_created)
        evt_get = self.coreutil_plugin.get_event(self.context,
                                                 evt_created['id'])
        self.assertEqual(evt_created['resource_id'], evt_get['resource_id'])
        self.assertEqual(evt_created['resource_state'],
                         evt_get['resource_state'])
        self.assertEqual(evt_created['resource_type'],
                         evt_get['resource_type'])
        self.assertEqual(evt_created['event_type'], evt_get['event_type'])
        self.assertEqual(evt_created['event_details'],
                         evt_get['event_details'])
        self.assertEqual(evt_created['timestamp'], evt_get['timestamp'])

    def test_get_events(self):
        evt_obj = self._get_dummy_event_obj()
        self.event_db_plugin.create_event(self.context,
                                          evt_obj['resource_id'],
                                          evt_obj['resource_type'],
                                          evt_obj['resource_state'],
                                          evt_obj['event_type'],
                                          evt_obj['timestamp'],
                                          evt_obj['event_details'])
        result = self.coreutil_plugin.get_events(self.context)
        self.assertTrue(len(result))

    def test_get_events_filtered_invalid_id(self):
        evt_obj = self._get_dummy_event_obj()
        self.event_db_plugin.create_event(self.context,
                                          evt_obj['resource_id'],
                                          evt_obj['resource_type'],
                                          evt_obj['resource_state'],
                                          evt_obj['event_type'],
                                          evt_obj['timestamp'],
                                          evt_obj['event_details'])
        result = self.coreutil_plugin.get_events(self.context, {'id': 'xyz'})
        self.assertFalse(len(result))

    def test_get_events_filtered_valid_id(self):
        evt_obj = self._get_dummy_event_obj()
        self.event_db_plugin.create_event(self.context,
                                          evt_obj['resource_id'],
                                          evt_obj['resource_type'],
                                          evt_obj['resource_state'],
                                          evt_obj['event_type'],
                                          evt_obj['timestamp'],
                                          evt_obj['event_details'])
        result = self.coreutil_plugin.get_events(self.context, {'id': '1'})
        self.assertTrue(len(result))

    def test_get_events_valid_fields(self):
        evt_obj = self._get_dummy_event_obj()
        self.event_db_plugin.create_event(self.context,
                                          evt_obj['resource_id'],
                                          evt_obj['resource_type'],
                                          evt_obj['resource_state'],
                                          evt_obj['event_type'],
                                          evt_obj['timestamp'],
                                          evt_obj['event_details'])
        result = self.coreutil_plugin.get_events(self.context, {'id': '1'},
                                            ['id', 'event_type'])
        self.assertTrue(len(result))
        self.assertIn('id', result[0])
        self.assertNotIn('resource_id', result[0])
        self.assertNotIn('resource_state', result[0])
        self.assertNotIn('resource_type', result[0])
        self.assertIn('event_type', result[0])
        self.assertNotIn('event_details', result[0])
        self.assertNotIn('timestamp', result[0])



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import oslo_messaging

from tacker.common import rpc
from tacker.sol_refactored.objects import base as objects_base


TOPIC_CONDUCTOR_V2 = 'TACKER_CONDUCTOR_V2'


class VnfLcmRpcApiV2(object):

    target = oslo_messaging.Target(
        exchange='tacker',
        topic=TOPIC_CONDUCTOR_V2,
        fanout=False,
        version='1.0')

    def _cast_lcm_op(self, context, lcmocc_id, method):
        serializer = objects_base.TackerObjectSerializer()

        client = rpc.get_client(self.target, version_cap=None,
                                serializer=serializer)
        cctxt = client.prepare()
        cctxt.cast(context, method, lcmocc_id=lcmocc_id)

    def start_lcm_op(self, context, lcmocc_id):
        self._cast_lcm_op(context, lcmocc_id, 'start_lcm_op')

    def retry_lcm_op(self, context, lcmocc_id):
        self._cast_lcm_op(context, lcmocc_id, 'retry_lcm_op')

    def rollback_lcm_op(self, context, lcmocc_id):
        self._cast_lcm_op(context, lcmocc_id, 'rollback_lcm_op')

    def modify_vnfinfo(self, context, lcmocc_id):
        self._cast_lcm_op(context, lcmocc_id, 'modify_vnfinfo')



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import fields


# NFV-SOL 005
# - v2.7.1 9.5.4.3 (API version: 2.0.0)
# - v2.8.1 9.5.4.3 (API version: 2.1.0)
# - v3.3.1 9.5.4.3 (API version: 2.1.0)
class PackageOnboardingStateType(fields.BaseTackerEnum):
    CREATED = 'CREATED'
    UPLOADING = 'UPLOADING'
    PROCESSING = 'PROCESSING'
    ONBOARDED = 'ONBOARDED'
    ERROR = 'ERROR'

    ALL = (CREATED, UPLOADING, PROCESSING, ONBOARDED, ERROR)


class PackageOnboardingStateTypeField(fields.BaseEnumField):
    AUTO_TYPE = PackageOnboardingStateType()


# NFV-SOL 005
# - v2.7.1 9.5.4.4 (API version: 2.0.0)
# - v2.8.1 9.5.4.4 (API version: 2.1.0)
# - v3.3.1 9.5.4.4 (API version: 2.1.0)
class PackageOperationalStateType(fields.BaseTackerEnum):
    ENABLED = 'ENABLED'
    DISABLED = 'DISABLED'

    ALL = (ENABLED, DISABLED)


class PackageOperationalStateTypeField(fields.BaseEnumField):
    AUTO_TYPE = PackageOperationalStateType()


# NFV-SOL 005
# - v2.7.1 9.5.4.5 (API version: 2.0.0)
# - v2.8.1 9.5.4.5 (API version: 2.1.0)
# - v3.3.1 9.5.4.5 (API version: 2.1.0)
class PackageUsageStateType(fields.BaseTackerEnum):
    IN_USE = 'IN_USE'
    NOT_IN_USE = 'NOT_IN_USE'

    ALL = (IN_USE, NOT_IN_USE)


class PackageUsageStateTypeField(fields.BaseEnumField):
    AUTO_TYPE = PackageUsageStateType()


# NFV-SOL 005
# - v2.7.1 9.5.4.6 (API version: 2.0.0)
# - v2.8.1 9.5.4.6 (API version: 2.1.0)
# - v3.3.1 9.5.4.6 (API version: 2.1.0)
class PackageChangeType(fields.BaseTackerEnum):
    OP_STATE_CHANGE = 'OP_STATE_CHANGE'
    PKG_DELETE = 'PKG_DELETE'

    ALL = (OP_STATE_CHANGE, PKG_DELETE)


class PackageChangeTypeField(fields.BaseEnumField):
    AUTO_TYPE = PackageChangeType()


# NFV-SOL 003
# - v3.3.1 5.5.4.3 (API version: 2.0.0)
class VnfOperationalStateType(fields.BaseTackerEnum):
    STARTED = 'STARTED'
    STOPPED = 'STOPPED'

    ALL = (STARTED, STOPPED)


class VnfOperationalStateTypeField(fields.BaseEnumField):
    AUTO_TYPE = VnfOperationalStateType()


# NFV-SOL 003
# - v3.3.1 5.5.4.4 (API version: 2.0.0)
class StopType(fields.BaseTackerEnum):
    FORCEFUL = 'FORCEFUL'
    GRACEFUL = 'GRACEFUL'

    ALL = (FORCEFUL, GRACEFUL)


class StopTypeField(fields.BaseEnumField):
    AUTO_TYPE = StopType()


# NFV-SOL 003
# - v3.3.1 5.5.4.5 (API version: 2.0.0)
class LcmOperationStateType(fields.BaseTackerEnum):
    STARTING = 'STARTING'
    PROCESSING = 'PROCESSING'
    COMPLETED = 'COMPLETED'
    FAILED_TEMP = 'FAILED_TEMP'
    FAILED = 'FAILED'
    ROLLING_BACK = 'ROLLING_BACK'
    ROLLED_BACK = 'ROLLED_BACK'

    ALL = (STARTING, PROCESSING, COMPLETED, FAILED_TEMP, FAILED,
           ROLLING_BACK, ROLLED_BACK)


class LcmOperationStateTypeField(fields.BaseEnumField):
    AUTO_TYPE = LcmOperationStateType()


# NFV-SOL 003
# - v3.3.1 5.5.4.6 (API version: 2.0.0)
class CancelModeType(fields.BaseTackerEnum):
    GRACEFUL = 'GRACEFUL'
    FORCEFUL = 'FORCEFUL'

    ALL = (GRACEFUL, FORCEFUL)


class CancelModeTypeField(fields.BaseEnumField):
    AUTO_TYPE = CancelModeType()


# NFV-SOL 003
# - v3.3.1 5.5.4.7 (API version: 2.0.0)
class LcmOperationType(fields.BaseTackerEnum):
    INSTANTIATE = 'INSTANTIATE'
    SCALE = 'SCALE'
    SCALE_TO_LEVEL = 'SCALE_TO_LEVEL'
    CHANGE_FLAVOUR = 'CHANGE_FLAVOUR'
    TERMINATE = 'TERMINATE'
    HEAL = 'HEAL'
    OPERATE = 'OPERATE'
    CHANGE_EXT_CONN = 'CHANGE_EXT_CONN'
    MODIFY_INFO = 'MODIFY_INFO'
    CREATE_SNAPSHOT = 'CREATE_SNAPSHOT'
    REVERT_TO_SNAPSHOT = 'REVERT_TO_SNAPSHOT'
    CHANGE_VNFPKG = 'CHANGE_VNFPKG'

    ALL = (INSTANTIATE, SCALE, SCALE_TO_LEVEL, CHANGE_FLAVOUR,
           TERMINATE, HEAL, OPERATE, CHANGE_EXT_CONN, MODIFY_INFO,
           CREATE_SNAPSHOT, REVERT_TO_SNAPSHOT, CHANGE_VNFPKG)


class LcmOperationTypeField(fields.BaseEnumField):
    AUTO_TYPE = LcmOperationType()


# NFV-SOL 003
# - v3.3.1 5.5.4.8 (API version: 2.0.0)
class LcmOpOccNotificationVerbosityType(fields.BaseTackerEnum):
    FULL = 'FULL'
    SHORT = 'SHORT'

    ALL = (FULL, SHORT)


class LcmOpOccNotificationVerbosityTypeField(fields.BaseEnumField):
    AUTO_TYPE = LcmOpOccNotificationVerbosityType()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import datetime
import iso8601
from tacker import context
from tacker.db.nfvo import nfvo_db
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnf(SqlTestCase):

    def setUp(self):
        super(TestVnf, self).setUp()
        self.context = context.get_admin_context()
        self.vnfd = self._create_vnfd()
        self.vims = self._create_vims()

    def _create_vnfd(self):
        vnfd_obj = objects.Vnfd(context=self.context, **fakes.vnfd_data)
        vnfd_obj.create()

        return vnfd_obj

    def _create_vims(self):
        vim_obj = nfvo_db.Vim(**fakes.vim_data)

        return vim_obj

    def test_save(self):
        vnf_data = fakes.get_vnf(self.vnfd.id, self.vims.id)
        vnf_obj = objects.vnf.VNF(context=self.context, **vnf_data)

        vnf_obj.id = uuidsentinel.instance_id
        vnf_obj.status = 'ERROR'
        vnf_obj.updated_at = datetime.datetime(
            1900, 1, 1, 1, 1, 1, tzinfo=iso8601.UTC)
        vnf_obj.save()

        self.assertEqual('ERROR', vnf_obj.status)

    @mock.patch('tacker.objects.vnf._vnf_get')
    def test_vnf_index_list(self, mock_vnf_get):
        vnf_data = fakes.get_vnf(self.vnfd.id, self.vims.id)
        vnf_obj = objects.vnf.VNF(context=self.context, **vnf_data)

        vnf_obj.id = uuidsentinel.instance_id
        vnf_obj.updated_at = datetime.datetime(
            1900, 1, 1, 1, 1, 1, tzinfo=iso8601.UTC)
        vnf_obj.save()

        mock_vnf_get.return_value = vnf_obj
        vnf_data_result = vnf_obj.vnf_index_list(vnf_obj.id, self.context)
        self.assertEqual('ACTIVE', vnf_data_result.status)



# Copyright 2012 New Dream Network, LLC (DreamHost)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# @author Mark McClain (DreamHost)

import sys
from unittest import mock


from tacker.db.migration import cli
from tacker.tests import base


class TestCli(base.BaseTestCase):
    def setUp(self):
        super(TestCli, self).setUp()
        self.do_alembic_cmd_p = mock.patch.object(cli, 'do_alembic_command')
        self.do_alembic_cmd = self.do_alembic_cmd_p.start()
        self.mock_alembic_err = mock.patch('alembic.util.err').start()
        self.mock_alembic_err.side_effect = SystemExit

    def _main_test_helper(self, argv, func_name, exp_args=(), exp_kwargs={}):
        with mock.patch.object(sys, 'argv', argv):
            cli.main()
            self.do_alembic_cmd.assert_has_calls(
                [mock.call(mock.ANY, func_name, *exp_args, **exp_kwargs)]
            )

    def test_stamp(self):
        self._main_test_helper(
            ['prog', 'stamp', 'foo'],
            'stamp',
            ('foo',),
            {'sql': False}
        )

        self._main_test_helper(
            ['prog', 'stamp', 'foo', '--sql'],
            'stamp',
            ('foo',),
            {'sql': True}
        )

    def test_current(self):
        self._main_test_helper(['prog', 'current'], 'current')

    def test_history(self):
        self._main_test_helper(['prog', 'history'], 'history')

    def test_check_migration(self):
        with mock.patch.object(cli, 'validate_head_file') as validate:
            self._main_test_helper(['prog', 'check_migration'], 'branches')
            validate.assert_called_once_with(mock.ANY)

    def test_database_sync_revision(self):
        with mock.patch.object(cli, 'update_head_file') as update:
            self._main_test_helper(
                ['prog', 'revision', '--autogenerate', '-m', 'message'],
                'revision',
                (),
                {'message': 'message', 'sql': False, 'autogenerate': True}
            )
            update.assert_called_once_with(mock.ANY)

            update.reset_mock()
            self._main_test_helper(
                ['prog', 'revision', '--sql', '-m', 'message'],
                'revision',
                (),
                {'message': 'message', 'sql': True, 'autogenerate': False}
            )
            update.assert_called_once_with(mock.ANY)

    def test_upgrade(self):
        self._main_test_helper(
            ['prog', 'upgrade', '--sql', 'head'],
            'upgrade',
            ('head',),
            {'sql': True}
        )

        self._main_test_helper(
            ['prog', 'upgrade', '--delta', '3'],
            'upgrade',
            ('+3',),
            {'sql': False}
        )

    def _test_validate_head_file_helper(self, heads, file_content=None):
        with mock.patch('alembic.script.ScriptDirectory.from_config') as fc:
            fc.return_value.get_heads.return_value = heads
            fc.return_value.get_current_head.return_value = heads[0]
            with mock.patch('builtins.open') as mock_open:
                mock_open.return_value.__enter__ = lambda s: s
                mock_open.return_value.__exit__ = mock.Mock()
                mock_open.return_value.read.return_value = file_content

                with mock.patch('os.path.isfile') as is_file:
                    is_file.return_value = file_content is not None

                    if file_content in heads:
                        cli.validate_head_file(mock.sentinel.config)
                    else:
                        self.assertRaises(
                            SystemExit,
                            cli.validate_head_file,
                            mock.sentinel.config
                        )
                        self.mock_alembic_err.assert_called_once_with(mock.ANY)
            fc.assert_called_once_with(mock.sentinel.config)

    def test_validate_head_file_multiple_heads(self):
        self._test_validate_head_file_helper(['a', 'b'])

    def test_validate_head_file_missing_file(self):
        self._test_validate_head_file_helper(['a'])

    def test_validate_head_file_wrong_contents(self):
        self._test_validate_head_file_helper(['a'], 'b')

    def test_validate_head_success(self):
        self._test_validate_head_file_helper(['a'], 'a')

    def test_update_head_file_multiple_heads(self):
        with mock.patch('alembic.script.ScriptDirectory.from_config') as fc:
            fc.return_value.get_heads.return_value = ['a', 'b']
            self.assertRaises(
                SystemExit,
                cli.update_head_file,
                mock.sentinel.config
            )
            self.mock_alembic_err.assert_called_once_with(mock.ANY)
            fc.assert_called_once_with(mock.sentinel.config)

    def test_update_head_file_success(self):
        with mock.patch('alembic.script.ScriptDirectory.from_config') as fc:
            fc.return_value.get_heads.return_value = ['a']
            fc.return_value.get_current_head.return_value = 'a'
            with mock.patch('builtins.open') as mock_open:
                mock_open.return_value.__enter__ = lambda s: s
                mock_open.return_value.__exit__ = mock.Mock()

                cli.update_head_file(mock.sentinel.config)
                mock_open.return_value.write.assert_called_once_with('a')
            fc.assert_called_once_with(mock.sentinel.config)



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.common import log
from tacker.db.common_services import common_services_db_plugin


class CommonServicesPlugin(common_services_db_plugin.CommonServicesPluginDb):
    """Reference plugin for COMMONSERVICES extension

    Implements the COMMONSERVICES extension and defines public facing APIs for
    common utility operations.
    """

    supported_extension_aliases = ['CommonServices']

    def __init__(self):
        super(CommonServicesPlugin, self).__init__()

    @log.log
    def get_event(self, context, event_id, fields=None):
        return super(CommonServicesPlugin, self).get_event(context, event_id,
                                                      fields)

    @log.log
    def get_events(self, context, filters=None, fields=None, sorts=None,
                   limit=None, marker_obj=None, page_reverse=False):
        return super(CommonServicesPlugin, self).get_events(context, filters,
                                                       fields, sorts, limit,
                                                       marker_obj,
                                                       page_reverse)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from tacker.api.vnflcm.v1 import sync_resource as sync
from tacker.conductor.conductorrpc import vnf_pkgm_rpc
from tacker import context
from tacker.db.nfvo import nfvo_db
from tacker.glance_store import store as glance_store
from tacker import objects
from tacker.tests.unit import base
from tacker.tests.unit.vnflcm import fakes as vnflcm_fakes
from tacker.tests.unit.vnfpkgm import fakes as vnfpkgm_fakes
from tacker.tests import uuidsentinel
import tacker.vnfm.nfvo_client as nfvo_client
from unittest import mock
from webob import exc


@ddt.ddt
class TestSyncVnfPackage(base.TestCase):

    def setUp(self):
        super(TestSyncVnfPackage, self).setUp()
        self.context = context.ContextBase(
            uuidsentinel.user_id,
            uuidsentinel.project_id,
            is_admin=True)
        self.vim = nfvo_db.Vim()

    def tearDown(self):
        super(TestSyncVnfPackage, self).tearDown()
        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(glance_store, 'store_csar')
    @mock.patch.object(objects.vnf_package_vnfd.VnfPackageVnfd,
                       'get_by_id')
    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, 'save')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    @mock.patch.object(vnf_pkgm_rpc.VNFPackageRPCAPI,
                       "upload_vnf_package_content")
    def test_package_and_package_vnfd_creation_successful(self,
            mock_upload_vnf_package_content,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_save_vnf_package,
            mock_vnf_package,
            mock_get_by_id,
            mock_glance_store):

        # glance_store mock Settings
        mock_glance_store.return_value = 'location', 0, 'checksum',\
                                         'multihash', 'loc_meta'

        updates = {'additionalArtifacts': [
            {
                'artifactPath': 'sample1_file.yaml',
                'checksum': {
                    'hash':
                        '53b504e608eef3d19a\
                            22413bf6ee72f42091fbb5213e6a876a3a22f6c3c94fe1',
                    'algorithm': 'SHA-256'
                }
            },
            {
                'artifactPath': 'sample2_file.yaml',
                'checksum': {
                    'hash':
                        '43b504e608eef3d19a\
                            22413bf6ee72f42091fbb5213e6a876a3a22f6c3c94fe1',
                    'algorithm': 'SHA-256'
                }
            }
        ]}
        vnf_package_info = vnfpkgm_fakes.index_response(
            remove_attrs=['userDefinedData'],
            vnf_package_updates=updates)[0]

        exp_vnf_package_vnfd = \
            vnflcm_fakes.fake_vnf_package_vnfd_model_dict(**updates)
        mock_get_by_id.return_value = exp_vnf_package_vnfd

        vnf_package_vnfd = sync.SyncVnfPackage.create_package(
            self.context, vnf_package_info)

        # Expected value setting
        self.assertEqual(exp_vnf_package_vnfd.get('package_uuid'),
         vnf_package_vnfd.get('package_uuid'))
        self.assertEqual(exp_vnf_package_vnfd.get('vnfd_id'),
         vnf_package_vnfd.get('vnfd_id'))
        self.assertEqual(exp_vnf_package_vnfd.get('vnf_provider'),
         vnf_package_vnfd.get('vnf_provider'))
        self.assertEqual(exp_vnf_package_vnfd.get('vnf_product_name'),
         vnf_package_vnfd.get('vnf_product_name'))
        self.assertEqual(exp_vnf_package_vnfd.get('vnf_software_version'),
         vnf_package_vnfd.get('vnf_software_version'))
        self.assertEqual(exp_vnf_package_vnfd.get('vnfd_version'),
         vnf_package_vnfd.get('vnfd_version'))

        # Check if a mock is called even once
        mock_upload_vnf_package_content.assert_called()
        mock_nfvo_download_vnf_packages.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_save_vnf_package.assert_called()
        mock_vnf_package.assert_called()
        mock_get_by_id.assert_called()
        mock_glance_store.assert_called()

    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    def test_package_and_package_vnfd_creation_package_create_err(self,
            mock_vnf_package):

        # vnf_package mock Settings
        mock_vnf_package.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_vnf_package.assert_called()

    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_undefinedexcep(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_vnf_package):

        # VnfPackageRequest.download_vnf_packages mock Settings
        mock_nfvo_download_vnf_packages.side_effect = \
            nfvo_client.UndefinedExternalSettingException(
                "Vnf package the external setting to 'base_url' undefined.")

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPNotFound,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_vnf_package.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_nfvo_download_vnf_packages.assert_called()

    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_falieddownloadexcep(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_vnf_package):

        # VnfPackageRequest.download_vnf_packages mock Settings
        vnf_package_zip = ''
        mock_nfvo_download_vnf_packages.side_effect = \
            nfvo_client.FaliedDownloadContentException(
                "Failed response content, vnf_package_zip={}".format(
                    vnf_package_zip))

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_vnf_package.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_nfvo_download_vnf_packages.assert_called()

    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_exception(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_vnf_package):

        # VnfPackageRequest.download_vnf_packages mock Settings
        mock_nfvo_download_vnf_packages.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_vnf_package.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_nfvo_download_vnf_packages.assert_called()

    @mock.patch.object(glance_store, 'store_csar')
    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_store_csar_err(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_vnf_package,
            mock_glance_store):

        # glance_store mock Settings
        mock_glance_store.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_vnf_package.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_nfvo_download_vnf_packages.assert_called()
        mock_glance_store.assert_called()

    @mock.patch.object(glance_store, 'store_csar')
    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, 'save')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_package_upde_err(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_save_vnf_package,
            mock_vnf_package,
            mock_glance_store):

        # glance_store mock Settings
        mock_glance_store.return_value = 'location', 0, 'checksum',\
                                         'multihash', 'loc_meta'

        # vnf_package mock Settings
        mock_save_vnf_package.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_nfvo_download_vnf_packages.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_save_vnf_package.assert_called()
        mock_vnf_package.assert_called()
        mock_glance_store.assert_called()

    @mock.patch.object(glance_store, 'store_csar')
    @mock.patch.object(vnf_pkgm_rpc.VNFPackageRPCAPI,
                       'upload_vnf_package_content')
    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, 'save')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_package_vnfd_creation_rpc_err(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_save_vnf_package,
            mock_vnf_package,
            mock_upload_vnf_package_content,
            mock_glance_store):

        # glance_store mock Settings
        mock_glance_store.return_value = 'location', 0, 'checksum',\
                                         'multihash', 'loc_meta'
        mock_upload_vnf_package_content.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_nfvo_download_vnf_packages.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_save_vnf_package.assert_called()
        mock_vnf_package.assert_called()
        mock_upload_vnf_package_content.assert_called()
        mock_glance_store.assert_called()

    @mock.patch.object(glance_store, 'store_csar')
    @mock.patch.object(objects.vnf_package_vnfd.VnfPackageVnfd,
                       'get_by_id')
    @mock.patch.object(objects.vnf_package, '_vnf_package_create')
    @mock.patch.object(objects.vnf_package.VnfPackage, 'save')
    @mock.patch.object(objects.vnf_package.VnfPackage, '_from_db_object')
    @mock.patch.object(nfvo_client.VnfPackageRequest, 'download_vnf_packages')
    def test_package_and_get_package_vnfd_err(self,
            mock_nfvo_download_vnf_packages,
            mock_from_db_vnf_package,
            mock_save_vnf_package,
            mock_vnf_package,
            mock_get_by_id,
            mock_glance_store):

        # glance_store mock Settings
        mock_glance_store.return_value = 'location', 0, 'checksum',\
                                         'multihash', 'loc_meta'

        # vnf_package_vnfd mock Settings
        mock_get_by_id.side_effect = Exception

        # SyncVnfPackage.create_package
        vnf_package_info = vnfpkgm_fakes.index_response()[0]
        self.assertRaises(
            exc.HTTPInternalServerError,
            sync.SyncVnfPackage.create_package,
            self.context,
            vnf_package_info)

        # Check if a mock is called even once
        mock_nfvo_download_vnf_packages.assert_called()
        mock_from_db_vnf_package.assert_called()
        mock_save_vnf_package.assert_called()
        mock_vnf_package.assert_called()
        mock_get_by_id.assert_called()
        mock_glance_store.assert_called()



#    Copyright 2020 Distributed Cloud and Network (DCN)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import time
import yaml

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file


class VnfTestMaintenanceMonitor(base.BaseTackerTest):

    def _test_vnf_tosca_maintenance(self, vnfd_file, vnf_name):
        input_yaml = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)
        tosca_arg = {'vnfd': {'name': vnf_name,
                              'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd with tosca template
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertIsNotNone(vnfd_instance)

        # Create vnf with vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']
        vnf_arg = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name}}
        vnf_instance = self.client.create_vnf(body=vnf_arg)
        vnf_id = vnf_instance['vnf']['id']

        self.validate_vnf_instance(vnfd_instance, vnf_instance)

        def _wait_vnf_active_and_assert_vdu_count(vdu_count, scale_type=None):
            self.wait_until_vnf_active(
                vnf_id,
                constants.VNF_CIRROS_CREATE_TIMEOUT,
                constants.ACTIVE_SLEEP_TIME)

            vnf = self.client.show_vnf(vnf_id)['vnf']
            self.assertEqual(vdu_count, len(jsonutils.loads(
                vnf['mgmt_ip_address'])['VDU1']))

        def _verify_maintenance_attributes(vnf_dict):
            vnf_attrs = vnf_dict.get('attributes', {})
            maintenance_vdus = vnf_attrs.get('maintenance', '{}')
            maintenance_vdus = jsonutils.loads(maintenance_vdus)
            maintenance_url = vnf_attrs.get('maintenance_url', '')
            words = maintenance_url.split('/')

            self.assertEqual(len(maintenance_vdus.keys()), 2)
            self.assertEqual(len(words), 8)
            self.assertEqual(words[5], vnf_dict['id'])
            self.assertEqual(words[7], vnf_dict['tenant_id'])

            maintenance_urls = {}
            for vdu, access_key in maintenance_vdus.items():
                maintenance_urls[vdu] = maintenance_url + '/' + access_key
            return maintenance_urls

        def _verify_maintenance_alarm(url, project_id):
            aodh_client = self.aodh_http_client()
            alarm_query = {
                'and': [
                    {'=': {'project_id': project_id}},
                    {'=~': {'alarm_actions': url}}]}

            # Check alarm instance for MAINTENANCE_ALL
            alarm_url = 'v2/query/alarms'
            encoded_data = jsonutils.dumps(alarm_query)
            encoded_body = jsonutils.dumps({'filter': encoded_data})
            resp, response_body = aodh_client.do_request(alarm_url, 'POST',
                                                         body=encoded_body)
            self.assertEqual(len(response_body), 1)
            alarm_dict = response_body[0]
            self.assertEqual(url, alarm_dict.get('alarm_actions', [])[0])
            return response_body[0]

        def _verify_maintenance_actions(vnf_dict, alarm_dict):
            tacker_client = self.tacker_http_client()
            alarm_url = alarm_dict.get('alarm_actions', [])[0]
            tacker_url = '/%s' % alarm_url[alarm_url.find('v1.0'):]

            def _request_maintenance_action(state):
                alarm_body = _create_alarm_data(vnf_dict, alarm_dict, state)
                resp, response_body = tacker_client.do_request(
                    tacker_url, 'POST', body=alarm_body)

                time.sleep(constants.SCALE_SLEEP_TIME)
                target_scaled = -1
                if state == 'MAINTENANCE':
                    target_scaled = 0
                elif state == 'SCALE_IN':
                    target_scaled = 1
                    _wait_vnf_active_and_assert_vdu_count(2, scale_type='in')
                elif state == 'MAINTENANCE_COMPLETE':
                    target_scaled = 0
                    _wait_vnf_active_and_assert_vdu_count(3, scale_type='out')

                updated_vnf = self.client.show_vnf(vnf_id)['vnf']
                scaled = updated_vnf['attributes'].get('maintenance_scaled',
                                                       '-1')
                self.assertEqual(int(scaled), target_scaled)
                time.sleep(constants.SCALE_WINDOW_SLEEP_TIME)

            _request_maintenance_action('MAINTENANCE')
            _request_maintenance_action('SCALE_IN')
            _request_maintenance_action('MAINTENANCE_COMPLETE')

            self.verify_vnf_crud_events(
                vnf_id, evt_constants.RES_EVT_SCALE,
                evt_constants.ACTIVE, cnt=2)
            self.verify_vnf_crud_events(
                vnf_id, evt_constants.RES_EVT_SCALE,
                evt_constants.PENDING_SCALE_OUT, cnt=1)
            self.verify_vnf_crud_events(
                vnf_id, evt_constants.RES_EVT_SCALE,
                evt_constants.PENDING_SCALE_IN, cnt=1)

        def _create_alarm_data(vnf_dict, alarm_dict, state):
            '''This function creates a raw payload of alarm to trigger Tacker directly.

            This function creates a raw payload which Fenix will put
            when Fenix process maintenance procedures. Alarm_receiver and
            specific steps of Fenix workflow will be tested by sending the raw
            to Tacker directly.
            '''
            utc_time = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
            fake_url = 'http://localhost/'
            sample_data = {
                'alarm_name': alarm_dict['name'],
                'alarm_id': alarm_dict['alarm_id'],
                'severity': 'low',
                'previous': 'alarm',
                'current': 'alarm',
                'reason': 'Alarm test for Tacker functional test',
                'reason_data': {
                    'type': 'event',
                    'event': {
                        'message_id': uuidutils.generate_uuid(),
                        'event_type': 'maintenance.scheduled',
                        'generated': utc_time,
                        'traits': [
                            ['project_id', 1, vnf_dict['tenant_id']],
                            ['allowed_actions', 1, '[]'],
                            ['instance_ids', 1, fake_url],
                            ['reply_url', 1, fake_url],
                            ['state', 1, state],
                            ['session_id', 1, uuidutils.generate_uuid()],
                            ['actions_at', 4, utc_time],
                            ['reply_at', 4, utc_time],
                            ['metadata', 1, '{}']
                        ],
                        'raw': {},
                        'message_signature': uuidutils.generate_uuid()
                    }
                }
            }
            return jsonutils.dumps(sample_data)

        _wait_vnf_active_and_assert_vdu_count(3)
        urls = _verify_maintenance_attributes(vnf_instance['vnf'])

        maintenance_url = urls.get('ALL', '')
        project_id = vnf_instance['vnf']['tenant_id']
        alarm_dict = _verify_maintenance_alarm(maintenance_url, project_id)
        _verify_maintenance_actions(vnf_instance['vnf'], alarm_dict)

        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, (
                'Failed to delete vnf %s after the maintenance test' % vnf_id)
        self.addCleanup(self.client.delete_vnfd, vnfd_id)
        self.addCleanup(self.wait_until_vnf_delete, vnf_id,
                        constants.VNF_CIRROS_DELETE_TIMEOUT)

    # TODO(kden): add a test for migration
    def test_vnf_alarm_maintenance(self):
        self._test_vnf_tosca_maintenance(
            'sample-tosca-vnfd-maintenance.yaml',
            'maintenance_vnf')



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import glance_store
import urllib
from urllib import error as urlerr

from tacker.common import exceptions
import tacker.conf
from tacker.glance_store import store
from tacker.tests.unit import base


CONF = tacker.conf.CONF


class StoreBaseTest(base.TestCase):

    def setUp(self):
        super(StoreBaseTest, self).setUp()
        self.conf = CONF
        glance_store.create_stores(self.conf)
        self.body = {"address_information": "http://welcome.com/test.zip",
                     "user_name": "user1", "password": "pass1"}

    @mock.patch.object(urllib.request, 'urlopen')
    def test_get_csar_data_iter_with_username_password(self, mock_url_open):
        store.get_csar_data_iter(self.body)
        mock_url_open.assert_called_once()

    @mock.patch.object(urllib.request, 'urlopen')
    def test_get_csar_data_iter_without_username_password(self, mock_url_open):
        body = {"address_information": "http://welcome.com/test.zip",
                "user_name": None, "password": None}
        store.get_csar_data_iter(body)
        mock_url_open.assert_called_once()

    @mock.patch.object(urllib.request, 'urlopen')
    def test_get_csar_data_iter_unauthorised(self, mock_url_open):
        mock_url_open.side_effect = urlerr.HTTPError(
            url='', code=401, msg='HTTP Error 401 Unauthorized', hdrs={},
            fp=None)
        self.assertRaises(exceptions.VNFPackageURLInvalid,
                          store.get_csar_data_iter, self.body)



# Copyright 2015 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import yaml

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file


class VnfmTestParam(base.BaseTackerTest):
    def _test_vnfd_create(self, vnfd_file, vnfd_name):
        yaml_input = read_file(vnfd_file)
        tosca_dict = yaml.safe_load(yaml_input)
        req_dict = {'vnfd': {'name': vnfd_name,
                    'attributes': {'vnfd': tosca_dict}}}

        # Create vnfd
        vnfd_instance = self.client.create_vnfd(body=req_dict)
        self.assertIsNotNone(vnfd_instance)
        vnfd_id = vnfd_instance['vnfd']['id']
        self.assertIsNotNone(vnfd_id)
        self.verify_vnfd_events(
            vnfd_id, evt_constants.RES_EVT_CREATE,
            evt_constants.RES_EVT_ONBOARDED)
        return vnfd_instance

    def _test_vnfd_delete(self, vnfd_instance):
        # Delete vnfd
        vnfd_id = vnfd_instance['vnfd']['id']
        self.assertIsNotNone(vnfd_id)
        try:
            self.client.delete_vnfd(vnfd_id)
        except Exception:
            assert False, "vnfd Delete failed"
        self.verify_vnfd_events(vnfd_id, evt_constants.RES_EVT_DELETE,
                                evt_constants.RES_EVT_NA_STATE)
        try:
            vnfd_d = self.client.show_vnfd(vnfd_id)
        except Exception:
            assert True, "Vnfd Delete success" + str(vnfd_d) + str(Exception)

    def _test_vnf_create(self, vnfd_instance, vnf_name, param_values):
        # Create the vnf with values
        vnfd_id = vnfd_instance['vnfd']['id']
        # Create vnf with values file
        vnf_dict = dict()
        vnf_dict = {'vnf': {'vnfd_id': vnfd_id, 'name': vnf_name,
                    'attributes': {'param_values': param_values}}}
        vnf_instance = self.client.create_vnf(body=vnf_dict)

        self.validate_vnf_instance(vnfd_instance, vnf_instance)
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.assertIsNotNone(self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address'])
        vnf_instance = self.client.show_vnf(vnf_id)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        # Verify values dictionary is same as param values from vnf_show

        param_values = vnf_instance['vnf']['attributes']['param_values']
        param_values_dict = yaml.safe_load(param_values)

        return vnf_instance, param_values_dict

    def _test_vnf_update(self, vnf_instance, param_values):
        # Update Vnf
        vnf_id = vnf_instance['vnf']['id']
        new_param_values = {'vnf': {'attributes': {
                            'param_values': param_values}}}
        self.client.update_vnf(vnf_id, new_param_values)
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_UPDATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)

        # Wait until the update on the heat side is completed,
        # because vnf deletion will cause a conflict without waiting for this.
        stack_id = self.client.show_vnf(vnf_id)['vnf']['instance_id']
        start_time = int(time.time())
        while True:
            vdu_resource = self.get_vdu_resource(stack_id, "VDU1")
            vdu_resource_dict = vdu_resource.to_dict()
            vdu_resource_status = vdu_resource_dict['resource_status']
            if ((int(time.time()) - start_time >
                    constants.VNF_CIRROS_UPDATE_TIMEOUT) or
                    (vdu_resource_status == 'UPDATE_COMPLETE')):
                break
            time.sleep(constants.ACTIVE_SLEEP_TIME)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_UPDATE, evt_constants.PENDING_UPDATE)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_UPDATE, evt_constants.ACTIVE)

        # Verify vnf_param_values_dict is same as param values from vnf_show
        vnf_instance = self.client.show_vnf(vnf_id)
        vnf_param_values = vnf_instance['vnf']['attributes']['param_values']
        vnf_param_values_dict = yaml.safe_load(vnf_param_values)

        # Verify stack_parameters is same as parameters from stack_show
        instance_id = vnf_instance['vnf']['instance_id']
        stack_values = self.h_client.stacks.get(instance_id)
        stack_parameters = stack_values.parameters

        return vnf_param_values_dict, stack_parameters

    def _test_vnf_delete(self, vnf_instance):
        # Delete Vnf
        vnf_id = vnf_instance['vnf']['id']
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"
        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

        try:
            vnf_d = self.client.show_vnf(vnf_id)
        except Exception:
            assert True, "Vnf Delete success" + str(vnf_d) + str(Exception)

    def test_vnfd_param_tosca_template(self):
        vnfd_name = 'sample_cirros_vnfd_tosca'
        vnfd_instance = self._test_vnfd_create(
            'sample-tosca-vnfd-param.yaml', vnfd_name)
        self._test_vnfd_delete(vnfd_instance)

    def test_vnf_param_tosca_template_default(self):
        self._test_vnf_param_tosca_template(
            vnfd_name='cirros_vnfd_tosca_param',
            vnfd_file='sample-tosca-vnfd-param.yaml',
            param_file='sample-tosca-vnf-values.yaml',
            vnf_name='test_vnf_with_parameters_tosca_template')

    def test_vnf_param_with_artifacts_image_tosca_template(self):
        self._test_vnf_param_tosca_template(
            vnfd_name='cirros_vnfd_tosca_param_artifacts_image',
            vnfd_file='sample-tosca-vnfd-param-artifacts-image.yaml',
            param_file='sample-tosca-vnf-artifacts-image-values.yaml',
            vnf_name='test_vnf_with_param_artifacts_image_tosca_templ')

    def _test_vnf_param_tosca_template(self, vnfd_file, vnfd_name,
                                       param_file, vnf_name):
        vnfd_instance = self._test_vnfd_create(vnfd_file, vnfd_name)

        # Get vnfd_id
        vnfd_id = vnfd_instance['vnfd']['id']

        # Add vnfd delete to cleanup job so that if vnf_instance fails to
        # create or update then it will be cleaned-up automatically
        # in tearDown()
        self.addCleanup(self.client.delete_vnfd, vnfd_id)

        # Create vnf instance
        values_str = read_file(param_file)
        values_dict = yaml.safe_load(values_str)
        vnf_instance, param_values_dict = self._test_vnf_create(
            vnfd_instance, vnf_name, values_dict)
        self.assertEqual(values_dict, param_values_dict)

        new_values_str = read_file('sample-tosca-vnf-update-values.yaml')
        new_values_dict = yaml.safe_load(new_values_str)
        vnf_param_values_dict, stack_parameters = self._test_vnf_update(
            vnf_instance, new_values_dict)
        for key, value in new_values_dict.items():
            if vnf_param_values_dict.get(key):
                self.assertEqual(value, vnf_param_values_dict[key])
            if stack_parameters.get(key):
                self.assertEqual(value, stack_parameters[key])

        self._test_vnf_delete(vnf_instance)
        vnf_id = vnf_instance['vnf']['id']
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)
        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import tempfile
import time
import urllib
import yaml

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
from tempest.lib import base

from tacker.sol_refactored.common import http_client
from tacker.sol_refactored.infra_drivers.openstack import heat_utils
from tacker.sol_refactored import objects
from tacker.tests.functional.common.fake_server import FakeServerManager
from tacker.tests.functional.sol_v2 import utils
from tacker.tests import utils as base_utils
from tacker import version

FAKE_SERVER_MANAGER = FakeServerManager()
MOCK_NOTIFY_CALLBACK_URL = '/notification/callback'

LOG = logging.getLogger(__name__)

VNF_PACKAGE_UPLOAD_TIMEOUT = 300


class BaseSolV2Test(base.BaseTestCase):
    """Base test case class for SOL v2 functional tests."""

    @classmethod
    def setUpClass(cls):
        super(BaseSolV2Test, cls).setUpClass()

        FAKE_SERVER_MANAGER.prepare_http_server()
        FAKE_SERVER_MANAGER.start_server()

        cfg.CONF(args=['--config-file', '/etc/tacker/tacker.conf'],
                 project='tacker',
                 version='%%prog %s' % version.version_info.release_string())
        objects.register_all()

        vim_info = cls.get_vim_info()
        cls.auth_url = vim_info.interfaceInfo['endpoint']

        auth = http_client.KeystonePasswordAuthHandle(
            auth_url=vim_info.interfaceInfo['endpoint'],
            username=vim_info.accessInfo['username'],
            password=vim_info.accessInfo['password'],
            project_name=vim_info.accessInfo['project'],
            user_domain_name=vim_info.accessInfo['userDomain'],
            project_domain_name=vim_info.accessInfo['projectDomain']
        )
        cls.tacker_client = http_client.HttpClient(auth)
        cls.neutron_client = http_client.HttpClient(auth,
                                                    service_type='network')
        cls.glance_client = http_client.HttpClient(auth,
                                                   service_type='image')
        cls.nova_client = http_client.HttpClient(auth,
                                                 service_type='compute')
        cls.heat_client = heat_utils.HeatClient(vim_info)

    @classmethod
    def tearDownClass(cls):
        super(BaseSolV2Test, cls).tearDownClass()
        FAKE_SERVER_MANAGER.stop_server()

    @classmethod
    def get_vim_info(cls):
        vim_params = yaml.safe_load(base_utils.read_file('local-vim.yaml'))
        vim_params['auth_url'] += '/v3'

        vim_info = objects.VimConnectionInfo(
            interfaceInfo={'endpoint': vim_params['auth_url']},
            accessInfo={
                'region': 'RegionOne',
                'project': vim_params['project_name'],
                'username': vim_params['username'],
                'password': vim_params['password'],
                'userDomain': vim_params['user_domain_name'],
                'projectDomain': vim_params['project_domain_name']
            }
        )

        return vim_info

    @classmethod
    def create_vnf_package(cls, sample_path, user_data={}, image_path=None):
        vnfd_id = uuidutils.generate_uuid()
        tmp_dir = tempfile.mkdtemp()

        utils.make_zip(sample_path, tmp_dir, vnfd_id, image_path)

        zip_file_name = os.path.basename(os.path.abspath(sample_path)) + ".zip"
        zip_file_path = os.path.join(tmp_dir, zip_file_name)

        path = "/vnfpkgm/v1/vnf_packages"
        req_body = {'userDefinedData': user_data}
        resp, body = cls.tacker_client.do_request(
            path, "POST", expected_status=[201], body=req_body)

        pkg_id = body['id']

        with open(zip_file_path, 'rb') as fp:
            path = "/vnfpkgm/v1/vnf_packages/{}/package_content".format(pkg_id)
            resp, body = cls.tacker_client.do_request(
                path, "PUT", body=fp, content_type='application/zip',
                expected_status=[202])

        # wait for onboard
        timeout = VNF_PACKAGE_UPLOAD_TIMEOUT
        start_time = int(time.time())
        path = "/vnfpkgm/v1/vnf_packages/{}".format(pkg_id)
        while True:
            resp, body = cls.tacker_client.do_request(
                path, "GET", expected_status=[200])
            if body['onboardingState'] == "ONBOARDED":
                break

            if ((int(time.time()) - start_time) > timeout):
                raise Exception("Failed to onboard vnf package")

            time.sleep(5)

        shutil.rmtree(tmp_dir)

        return pkg_id, vnfd_id

    @classmethod
    def delete_vnf_package(cls, pkg_id):
        path = "/vnfpkgm/v1/vnf_packages/{}".format(pkg_id)
        req_body = {"operationalState": "DISABLED"}
        resp, _ = cls.tacker_client.do_request(
            path, "PATCH", body=req_body)
        if resp.status_code != 200:
            LOG.error("failed to set operationalState to DISABLED")
            return

        cls.tacker_client.do_request(path, "DELETE")

    def setUp(self):
        super().setUp()

        callback_url = os.path.join(
            MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        FAKE_SERVER_MANAGER.clear_history(callback_url)
        FAKE_SERVER_MANAGER.set_callback('POST', callback_url, status_code=204)
        FAKE_SERVER_MANAGER.set_callback('GET', callback_url, status_code=204)

    def get_vnf_package(self, pkg_id):
        path = "/vnfpkgm/v1/vnf_packages/{}".format(pkg_id)
        resp, body = self.tacker_client.do_request(path, "GET")
        return body

    def get_network_ids(self, networks):
        path = "/v2.0/networks"
        resp, body = self.neutron_client.do_request(path, "GET")
        net_ids = {}
        for net in body['networks']:
            if net['name'] in networks:
                net_ids[net['name']] = net['id']
        return net_ids

    def get_subnet_ids(self, subnets):
        path = "/v2.0/subnets"
        resp, body = self.neutron_client.do_request(path, "GET")
        subnet_ids = {}
        for subnet in body['subnets']:
            if subnet['name'] in subnets:
                subnet_ids[subnet['name']] = subnet['id']
        return subnet_ids

    def create_network(self, name):
        path = "/v2.0/networks"
        req_body = {
            "network": {
                # "admin_state_up": true,
                "name": name
            }
        }
        try:
            resp, resp_body = self.neutron_client.do_request(
                path, "POST", body=req_body)
            return resp_body['network']['id']
        except Exception as e:
            self.fail("Failed, create network for name=<%s>, %s" %
                (name, e))

    def delete_network(self, net_id):
        path = "/v2.0/networks/{}".format(net_id)
        try:
            self.neutron_client.do_request(path, "DELETE")
        except Exception as e:
            self.fail("Failed, delete network for id=<%s>, %s" %
                (net_id, e))

    def create_subnet(self, net_id, sub_name, sub_range, ip_version):
        path = "/v2.0/subnets"
        req_body = {
            "subnet": {
                "name": sub_name,
                "network_id": net_id,
                "cidr": sub_range,
                "ip_version": ip_version
            }
        }
        try:
            resp, resp_body = self.neutron_client.do_request(
                path, "POST", body=req_body)
            return resp_body['subnet']['id']
        except Exception as e:
            self.fail("Failed, create subnet for name=<%s>, %s" %
                (sub_name, e))

    def delete_subnet(self, sub_id):
        path = "/v2.0/subnets/{}".format(sub_id)
        try:
            self.neutron_client.do_request(path, "DELETE")
        except Exception as e:
            self.fail("Failed, delete subnet for id=<%s>, %s" %
                (sub_id, e))

    def create_port(self, network_id, name):
        path = "/v2.0/ports"
        req_body = {
            'port': {
                'network_id': network_id,
                'name': name
            }
        }
        try:
            resp, resp_body = self.neutron_client.do_request(
                path, "POST", body=req_body)
            return resp_body['port']['id']
        except Exception as e:
            self.fail("Failed, create port for net_id=<%s>, %s" %
                (network_id, e))

    def delete_port(self, port_id):
        path = "/v2.0/ports/{}".format(port_id)
        try:
            self.neutron_client.do_request(path, "DELETE")
        except Exception as e:
            self.fail("Failed, delete port for id=<%s>, %s" %
                (port_id, e))

    def get_image_id(self, image_name):
        path = "/v2.0/images"
        resp, resp_body = self.glance_client.do_request(path, "GET")

        image_id = None
        for image in resp_body.get('images'):
            if image_name == image['name']:
                image_id = image['id']
        return image_id

    def get_server_details(self, server_name):
        path = "/servers/detail"
        resp, resp_body = self.nova_client.do_request(path, "GET")

        server_details = None
        for server in resp_body.get('servers'):
            if server_name == server['name']:
                server_details = server
        return server_details

    def create_vnf_instance(self, req_body):
        path = "/vnflcm/v2/vnf_instances"
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def delete_vnf_instance(self, inst_id):
        path = "/vnflcm/v2/vnf_instances/{}".format(inst_id)
        return self.tacker_client.do_request(
            path, "DELETE", version="2.0.0")

    def show_vnf_instance(self, inst_id):
        path = "/vnflcm/v2/vnf_instances/{}".format(inst_id)
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def list_vnf_instance(self, filter_expr=None):
        path = "/vnflcm/v2/vnf_instances"
        if filter_expr:
            path = "{}?{}".format(path, urllib.parse.urlencode(filter_expr))
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def instantiate_vnf_instance(self, inst_id, req_body):
        path = "/vnflcm/v2/vnf_instances/{}/instantiate".format(inst_id)
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def scale_vnf_instance(self, inst_id, req_body):
        path = f"/vnflcm/v2/vnf_instances/{inst_id}/scale"
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def update_vnf_instance(self, inst_id, req_body):
        path = f"/vnflcm/v2/vnf_instances/{inst_id}"
        return self.tacker_client.do_request(
            path, "PATCH", body=req_body, version="2.0.0")

    def heal_vnf_instance(self, inst_id, req_body):
        path = f"/vnflcm/v2/vnf_instances/{inst_id}/heal"
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def change_ext_conn(self, inst_id, req_body):
        path = f"/vnflcm/v2/vnf_instances/{inst_id}/change_ext_conn"
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def terminate_vnf_instance(self, inst_id, req_body):
        path = "/vnflcm/v2/vnf_instances/{}/terminate".format(inst_id)
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def wait_lcmocc_complete(self, lcmocc_id):
        # NOTE: It is not necessary to set timeout because the operation
        # itself set timeout and the state will become 'FAILED_TEMP'.
        path = "/vnflcm/v2/vnf_lcm_op_occs/{}".format(lcmocc_id)
        while True:
            time.sleep(5)
            _, body = self.tacker_client.do_request(
                path, "GET", expected_status=[200], version="2.0.0")
            state = body['operationState']
            if state == 'COMPLETED':
                return
            elif state in ['STARTING', 'PROCESSING']:
                continue
            else:  # FAILED_TEMP or ROLLED_BACK
                raise Exception("Operation failed. state: %s" % state)

    def wait_lcmocc_failed_temp(self, lcmocc_id):
        # NOTE: It is not necessary to set timeout because the operation
        # itself set timeout and the state will become 'FAILED_TEMP'.
        path = "/vnflcm/v2/vnf_lcm_op_occs/{}".format(lcmocc_id)
        while True:
            time.sleep(5)
            _, body = self.tacker_client.do_request(
                path, "GET", expected_status=[200], version="2.0.0")
            state = body['operationState']
            if state == 'FAILED_TEMP':
                return
            elif state in ['STARTING', 'PROCESSING']:
                continue
            elif state == 'COMPLETED':
                raise Exception("Operation unexpected COMPLETED.")
            else:  # ROLLED_BACK
                raise Exception("Operation failed. state: %s" % state)

    def wait_lcmocc_rolled_back(self, lcmocc_id):
        # NOTE: It is not necessary to set timeout because the operation
        # itself set timeout and the state will become 'FAILED_TEMP'.
        path = f"/vnflcm/v2/vnf_lcm_op_occs/{lcmocc_id}"
        while True:
            time.sleep(5)
            _, body = self.tacker_client.do_request(
                path, "GET", expected_status=[200], version="2.0.0")
            state = body['operationState']
            if state == 'ROLLED_BACK':
                return
            if state in ['ROLLING_BACK']:
                continue

            raise Exception(f"Operation failed. state: {state}")

    def show_lcmocc(self, lcmocc_id):
        path = "/vnflcm/v2/vnf_lcm_op_occs/{}".format(lcmocc_id)
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def list_lcmocc(self, filter_expr=None):
        path = "/vnflcm/v2/vnf_lcm_op_occs"
        if filter_expr:
            path = "{}?{}".format(path, urllib.parse.urlencode(filter_expr))
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def rollback_lcmocc(self, lcmocc_id):
        path = f"/vnflcm/v2/vnf_lcm_op_occs/{lcmocc_id}/rollback"
        return self.tacker_client.do_request(
            path, "POST", version="2.0.0")

    def retry_lcmocc(self, lcmocc_id):
        path = f"/vnflcm/v2/vnf_lcm_op_occs/{lcmocc_id}/retry"
        return self.tacker_client.do_request(
            path, "POST", version="2.0.0")

    def fail_lcmocc(self, lcmocc_id):
        path = f"/vnflcm/v2/vnf_lcm_op_occs/{lcmocc_id}/fail"
        return self.tacker_client.do_request(
            path, "POST", version="2.0.0")

    def create_subscription(self, req_body):
        path = "/vnflcm/v2/subscriptions"
        return self.tacker_client.do_request(
            path, "POST", body=req_body, version="2.0.0")

    def delete_subscription(self, sub_id):
        path = "/vnflcm/v2/subscriptions/{}".format(sub_id)
        return self.tacker_client.do_request(
            path, "DELETE", version="2.0.0")

    def show_subscription(self, sub_id):
        path = "/vnflcm/v2/subscriptions/{}".format(sub_id)
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def list_subscriptions(self, filter_expr=None):
        path = "/vnflcm/v2/subscriptions"
        if filter_expr:
            path = "{}?{}".format(path, urllib.parse.urlencode(filter_expr))
        return self.tacker_client.do_request(
            path, "GET", version="2.0.0")

    def _check_resp_headers(self, resp, supported_headers):
        unsupported_headers = ['Link', 'Retry-After',
                               'Content-Range', 'WWW-Authenticate']
        for s in supported_headers:
            if s not in resp.headers:
                raise Exception("Supported header doesn't exist: %s" % s)
        for u in unsupported_headers:
            if u in resp.headers:
                raise Exception("Unsupported header exist: %s" % u)

    def check_resp_headers_in_create(self, resp):
        # includes location header and response body
        supported_headers = ['Version', 'Location', 'Content-Type',
                             'Accept-Ranges']
        self._check_resp_headers(resp, supported_headers)

    def check_resp_headers_in_operation_task(self, resp):
        # includes location header and no response body
        supported_headers = ['Version', 'Location']
        self._check_resp_headers(resp, supported_headers)

    def check_resp_headers_in_get(self, resp):
        # includes response body and no location header
        supported_headers = ['Version', 'Content-Type',
                             'Accept-Ranges']
        self._check_resp_headers(resp, supported_headers)

    def check_resp_headers_in_delete(self, resp):
        # no location header and response body
        supported_headers = ['Version']
        self._check_resp_headers(resp, supported_headers)

    def check_resp_body(self, body, expected_attrs):
        for attr in expected_attrs:
            if attr not in body:
                raise Exception("Expected attribute doesn't exist: %s" % attr)

    def assert_notification_get(self, callback_url):
        notify_mock_responses = FAKE_SERVER_MANAGER.get_history(
            callback_url)
        FAKE_SERVER_MANAGER.clear_history(
            callback_url)
        self.assertEqual(1, len(notify_mock_responses))
        self.assertEqual(204, notify_mock_responses[0].status_code)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from tacker.common import log

LOG = logging.getLogger(__name__)


class HOTUpdater(object):
    """Update HOT template."""

    def __init__(self, heatclient):
        self.heatclient = heatclient
        self.template = {}
        self.nested_templates = dict()

    @log.log
    def get_templates_from_stack(self, stack_id):
        """Get template information from the stack.

        Get the template from stack specified by stack_id,
        if stack has scalable resource, get the its child
        template.
        """

        def _get_resource(name, resources):
            for resource in resources:
                if resource.resource_name == name:
                    return resource

        self.template = self.heatclient.stacks.template(stack_id)
        LOG.debug('got main template for stack({}). template={}'.format(
            stack_id, self.template))

        stack_resources = self.heatclient.resource_get_list(stack_id,
            nested_depth=2)
        for resource in stack_resources:
            if resource.resource_type == 'OS::Heat::AutoScalingGroup':
                intermediate_template = self.heatclient.stacks.template(
                    resource.physical_resource_id)

                for resource_id in intermediate_template['resources'].keys():
                    corresponding_resource = _get_resource(resource_id,
                        stack_resources)
                    nested_template = self.heatclient.stacks.template(
                        corresponding_resource.physical_resource_id)
                    LOG.debug('got nested template for stack({}). template={}'
                        .format(corresponding_resource.physical_resource_id,
                        nested_template))
                if nested_template:
                    self.nested_templates[
                        corresponding_resource.resource_type] = nested_template

    @log.log
    def update_resource_property(self,
                                resource_id,
                                resource_types=[],
                                **kwargs):
        """Update attributes of resource properties.

        Get the resource information from template's resources section,
        and update properties using kwargs information.
        If resource type does not include in resource_types, nothing to do.
        """

        def _update(template, resource_id, resource_types, kwargs):
            resource = template.get('resources', {}).get(resource_id)
            if not resource:
                return
            if resource.get('type', {}) not in resource_types:
                return

            resource_properties = resource.get('properties', {})
            if not resource_properties:
                return

            for key, value in kwargs.items():
                if value is not None:
                    resource_properties.update({key: value})
                elif resource_properties.get(key):
                    del resource_properties[key]

        _update(self.template, resource_id, resource_types, kwargs)

        for value in self.nested_templates.values():
            nested_template = value
            _update(nested_template, resource_id, resource_types, kwargs)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy

from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy import exc
from sqlalchemy.orm import joinedload

from tacker._i18n import _
from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.db.vnfm import vnfm_db
from tacker import objects
from tacker.objects import base
from tacker.objects import common
from tacker.objects import fields
from tacker.objects import vnf_instantiated_info
from tacker.objects import vnf_package as vnf_package_obj
from tacker.objects import vnf_package_vnfd


LOG = logging.getLogger(__name__)


@db_api.context_manager.reader
def _vnf_instance_get_by_id(context, vnf_instance_id, columns_to_join=None,
                            read_deleted="no"):

    query = api.model_query(context, models.VnfInstance,
                            read_deleted=read_deleted,
                            project_only=True). \
        filter_by(id=vnf_instance_id)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    result = query.first()

    if not result:
        raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

    return result


@db_api.context_manager.writer
def _vnf_instance_create(context, values):
    vnf_instance = models.VnfInstance()
    vnf_instance.update(values)
    vnf_instance.save(context.session)

    return _vnf_instance_get_by_id(context, vnf_instance.id,
                                   columns_to_join=["instantiated_vnf_info"])


@db_api.context_manager.writer
def _vnf_instance_update(context, vnf_instance_id, values,
                         columns_to_join=None):

    vnf_instance = _vnf_instance_get_by_id(context, vnf_instance_id,
                                           columns_to_join=columns_to_join)
    vnf_instance.update(values)
    vnf_instance.save(session=context.session)

    return vnf_instance


@db_api.context_manager.writer
def _destroy_vnf_instance(context, uuid):
    now = timeutils.utcnow()
    updated_values = {'deleted': True,
                      'deleted_at': now
                      }
    api.model_query(context, models.VnfInstantiatedInfo). \
        filter_by(vnf_instance_id=uuid). \
        update(updated_values, synchronize_session=False)

    api.model_query(context, models.VnfInstance).\
        filter_by(id=uuid). \
        update(updated_values, synchronize_session=False)


@db_api.context_manager.reader
def _vnf_instance_list(context, columns_to_join=None):
    query = api.model_query(context, models.VnfInstance, read_deleted="no",
                            project_only=True)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    return query.all()


@db_api.context_manager.reader
def _vnf_instance_list_by_filter(context, columns_to_join=None,
                                 filters=None):
    query = api.model_query(context, models.VnfInstance,
                            read_deleted="no",
                            project_only=True)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    if filters:
        query = common.apply_filters(query, filters)

    return query.all()


def _make_vnf_instance_list(context, vnf_instance_list, db_vnf_instance_list,
                            expected_attrs):
    vnf_instance_cls = VnfInstance

    vnf_instance_list.objects = []
    for db_vnf_instance in db_vnf_instance_list:
        vnf_instance_obj = vnf_instance_cls._from_db_object(
            context, vnf_instance_cls(context), db_vnf_instance,
            expected_attrs=expected_attrs)
        vnf_instance_list.objects.append(vnf_instance_obj)

    vnf_instance_list.obj_reset_changes()
    return vnf_instance_list


# decorator to catch DBAccess exception
def _wrap_object_error(method):

    def wrapper(*args, **kwargs):
        try:
            method(*args, **kwargs)
        except exc.SQLAlchemyError:
            raise exceptions.DBAccessError

    return wrapper


@db_api.context_manager.reader
def _get_vnf_instance(context, id):
    vnf_instance = api.model_query(
        context, models.VnfInstance).filter_by(
        vnfd_id=id).first()
    return vnf_instance


@db_api.context_manager.reader
def _vnf_instance_get(context, vnfd_id, columns_to_join=None):
    query = api.model_query(context, models.VnfInstance, read_deleted="no",
                            project_only=True).filter_by(vnfd_id=vnfd_id)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    return query.first()


def _merge_vim_connection_info(
        pre_vim_connection_info_list,
        update_vim_connection_info_list):

    def update_nested_element(pre_data, update_data):
        for key, val in update_data.items():
            if not isinstance(val, dict):
                pre_data[key] = val
                continue

            if key in pre_data:
                pre_data[key].update(val)
            else:
                pre_data.update({key: val})

    result = []
    clone_pre_list = copy.deepcopy(pre_vim_connection_info_list)

    for update_vim_connection in update_vim_connection_info_list:
        pre_data = None
        for i in range(0, len(clone_pre_list) - 1):
            if clone_pre_list[i].id == update_vim_connection.get('id'):
                pre_data = clone_pre_list.pop(i)

        if pre_data is None:
            # new elm.
            result.append(objects.VimConnectionInfo._from_dict(
                update_vim_connection))
            continue

        convert_dict = pre_data.to_dict()
        update_nested_element(convert_dict, update_vim_connection)
        result.append(objects.VimConnectionInfo._from_dict(
            convert_dict))

    # Reflecting unupdated data
    result.extend(clone_pre_list)

    return result


@db_api.context_manager.writer
def _update_vnf_instances(
        context,
        vnf_lcm_opoccs,
        body_data,
        vnfd_pkg_data,
        vnfd_id):
    updated_values = {}
    updated_values['vnf_instance_name'] = body_data.get('vnf_instance_name')
    updated_values['vnf_instance_description'] = body_data.get(
        'vnf_instance_description')

    # get vnf_instances
    vnf_instance = _get_vnf_instance(context, vnfd_id)
    if body_data.get('metadata'):
        vnf_instance.vnf_metadata.update(body_data.get('metadata'))
        updated_values['vnf_metadata'] = vnf_instance.vnf_metadata

    if body_data.get('vim_connection_info'):
        merge_vim_connection_info = _merge_vim_connection_info(
            vnf_instance.vim_connection_info,
            body_data.get('vim_connection_info'))

        updated_values['vim_connection_info'] = merge_vim_connection_info

    if vnfd_pkg_data and len(vnfd_pkg_data) > 0:
        updated_values['vnfd_id'] = vnfd_pkg_data.get('vnfd_id')
        updated_values['vnf_provider'] = vnfd_pkg_data.get('vnf_provider')
        updated_values['vnf_product_name'] = vnfd_pkg_data.get(
            'vnf_product_name')
        updated_values['vnf_software_version'] = vnfd_pkg_data.get(
            'vnf_software_version')
        updated_values['vnf_pkg_id'] = vnfd_pkg_data.get('package_uuid')

    api.model_query(context, models.VnfInstance). \
        filter_by(id=vnf_lcm_opoccs.get('vnf_instance_id')). \
        update(updated_values, synchronize_session=False)

    vnf_now = timeutils.utcnow()
    if (body_data.get('vnfd_id') or body_data.get('vnf_pkg_id')):
        # update vnf
        if body_data.get('vnfd_id'):
            updated_values = {'vnfd_id': body_data.get('vnfd_id'),
                              'updated_at': vnf_now
                              }
        elif body_data.get('vnf_pkg_id'):
            updated_values = {'vnfd_id': vnfd_pkg_data.get('vnfd_id'),
                              'updated_at': vnf_now
                              }
        api.model_query(context, vnfm_db.VNF).\
            filter_by(id=vnf_lcm_opoccs.get('vnf_instance_id')). \
            update(updated_values, synchronize_session=False)

        # get vnf_packages
        id = vnfd_pkg_data.get('package_uuid')
        try:
            vnf_package = vnf_package_obj.VnfPackage.get_by_id(context, id)
        except exceptions.VnfPackageNotFound:
            raise exceptions.VnfPackageNotFound(id=id)

        if vnf_package.usage_state == 'NOT_IN_USE':
            # update vnf_packages
            now = timeutils.utcnow()
            updated_values = {'usage_state': 'IN_USE',
                              'updated_at': now
                              }
            api.model_query(context, models.VnfPackage).\
                filter_by(id=id). \
                update(updated_values, synchronize_session=False)

        # get vnf_instances
        vnf_instance = _get_vnf_instance(context, vnfd_id)

        if not vnf_instance:
            # get vnf_package_vnfd
            vnfd_data = vnf_package_vnfd.VnfPackageVnfd.get_by_vnfdId(
                context, vnfd_id)

            # update vnf_packages
            now = timeutils.utcnow()
            updated_values = {'usage_state': 'NOT_IN_USE',
                              'updated_at': now
                              }
            api.model_query(context, models.VnfPackage).\
                filter_by(id=vnfd_data.package_uuid). \
                update(updated_values, synchronize_session=False)

    return vnf_now


@base.TackerObjectRegistry.register
class VnfInstance(base.TackerObject, base.TackerPersistentObject,
                  base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'vnf_instance_name': fields.StringField(nullable=True),
        'vnf_instance_description': fields.StringField(nullable=True),
        'instantiation_state':
        fields.VnfInstanceStateField(
            nullable=False,
            default=fields.VnfInstanceState.NOT_INSTANTIATED),
        'task_state': fields.StringField(nullable=True, default=None),
        'vnfd_id': fields.StringField(nullable=False),
        'vnf_provider': fields.StringField(nullable=False),
        'vnf_product_name': fields.StringField(nullable=False),
        'vnf_software_version': fields.StringField(nullable=False),
        'vnfd_version': fields.StringField(nullable=False),
        'vim_connection_info': fields.ListOfObjectsField(
            'VimConnectionInfo', nullable=True, default=[]),
        'tenant_id': fields.StringField(nullable=False),
        'vnf_metadata': fields.DictOfStringsField(nullable=True, default={}),
        'vnf_pkg_id': fields.StringField(nullable=False),
        'instantiated_vnf_info': fields.ObjectField('InstantiatedVnfInfo',
                                                nullable=True, default=None)
    }

    ALL_ATTRIBUTES = {
        'id': ('id', "uuid", 'VnfInstance'),
        'vnfInstanceName': ('vnf_instance_name', 'string', 'VnfInstance'),
        'vnfInstanceDescription': (
            'vnf_instance_description', 'string', 'VnfInstance'),
        'instantiationState': ('instantiation_state', 'string', 'VnfInstance'),
        'taskState': ('task_state', 'string', 'VnfInstance'),
        'vnfdId': ('vnfd_id', 'uuid', 'VnfInstance'),
        'vnfProvider': ('vnf_provider', 'string', 'VnfInstance'),
        'vnfProductName': ('vnf_product_name', 'string', 'VnfInstance'),
        'vnfSoftwareVersion': (
            'vnf_software_version', 'string', 'VnfInstance'),
        'vnfdVersion': ('vnfd_version', 'string', 'VnfInstance'),
        'tenantId': ('tenant_id', 'string', 'VnfInstance'),
        'vnfPkgId': ('vnf_pkg_id', 'uuid', 'VnfInstance'),
        'vimConnectionInfo/*': ('vim_connection_info', 'key_value_pair',
                                {"key_column": "key", "value_column": "value",
                                 "model": "VnfInstance"}),
        'metadata/*': ('vnf_metadata', 'key_value_pair',
                       {"key_column": "key", "value_column": "value",
                        "model": "VnfInstance"}),
    }

    ALL_ATTRIBUTES.update(
        vnf_instantiated_info.InstantiatedVnfInfo.ALL_ATTRIBUTES)

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())

    def __init__(self, context=None, **kwargs):
        super(VnfInstance, self).__init__(context, **kwargs)
        self.obj_set_defaults()

    @staticmethod
    def _from_db_object(context, vnf_instance, db_vnf_instance,
                        expected_attrs=None):

        special_fields = ["instantiated_vnf_info", "vim_connection_info"]
        for key in vnf_instance.fields:
            if key in special_fields:
                continue

            setattr(vnf_instance, key, db_vnf_instance[key])

        VnfInstance._load_instantiated_vnf_info_from_db_object(
            context, vnf_instance, db_vnf_instance)

        vim_connection_info = db_vnf_instance['vim_connection_info']
        vim_connection_list = [objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context) for vim_info in vim_connection_info]
        vnf_instance.vim_connection_info = vim_connection_list

        vnf_instance._context = context
        vnf_instance.obj_reset_changes()

        return vnf_instance

    @staticmethod
    def _load_instantiated_vnf_info_from_db_object(context, vnf_instance,
                                                   db_vnf_instance):
        if db_vnf_instance['instantiated_vnf_info']:
            inst_vnf_info = objects.InstantiatedVnfInfo.obj_from_db_obj(
                context, db_vnf_instance['instantiated_vnf_info'])
            vnf_instance.instantiated_vnf_info = inst_vnf_info

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='create',
                                               reason=_('already created'))
        updates = self.obj_get_changes()

        if 'id' not in updates:
            updates['id'] = uuidutils.generate_uuid()
            self.id = updates['id']

        # add default vnf_instance_name if not specified
        # format: 'vnf' + <vnf instance id>
        if 'vnf_instance_name' not in updates or \
                not updates.get("vnf_instance_name"):
            updates['vnf_instance_name'] = 'vnf-' + self.id
            self.vnf_instance_name = updates['vnf_instance_name']

        db_vnf_instance = _vnf_instance_create(self._context, updates)
        expected_attrs = ["instantiated_vnf_info"]
        self._from_db_object(self._context, self, db_vnf_instance,
                             expected_attrs=expected_attrs)

    @base.remotable
    @_wrap_object_error
    def save(self):
        context = self._context

        updates = {}
        changes = self.obj_what_changed()

        for field in self.fields:
            if (self.obj_attr_is_set(field) and
                    isinstance(self.fields[field], fields.ObjectField)):
                try:
                    getattr(self, '_save_%s' % field)(context)
                except AttributeError:
                    LOG.exception('No save handler for %s', field)
            elif (self.obj_attr_is_set(field) and
                    isinstance(self.fields[field], fields.ListOfObjectsField)):
                field_list = getattr(self, field)
                updates[field] = [obj.obj_to_primitive() for obj in field_list]
            elif field in changes:
                if (field == 'vnf_instance_name' and
                        not self[field]):
                    self.vnf_instance_name = 'vnf-' + self.id

                updates[field] = self[field]

        expected_attrs = ["instantiated_vnf_info"]
        db_vnf_instance = _vnf_instance_update(self._context,
                                               self.id, updates,
                                               columns_to_join=expected_attrs)
        self._from_db_object(self._context, self, db_vnf_instance)

    def _save_instantiated_vnf_info(self, context):
        if self.instantiated_vnf_info:
            with self.instantiated_vnf_info.obj_alternate_context(context):
                self.instantiated_vnf_info.save()

    @base.remotable
    @_wrap_object_error
    def update_metadata(self, data):
        _metadata = copy.deepcopy(self['vnf_metadata'])
        _metadata.update(data)
        self['vnf_metadata'] = _metadata
        self.save()

    @base.remotable
    @_wrap_object_error
    def destroy(self, context):
        if not self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='destroy',
                                               reason='no uuid')

        _destroy_vnf_instance(context, self.id)

    def to_dict(self):
        data = {'id': self.id,
            'vnf_instance_name': self.vnf_instance_name,
            'vnf_instance_description': self.vnf_instance_description,
            'instantiation_state': self.instantiation_state,
            'vnfd_id': self.vnfd_id,
            'vnf_provider': self.vnf_provider,
            'vnf_product_name': self.vnf_product_name,
            'vnf_software_version': self.vnf_software_version,
            'vnfd_version': self.vnfd_version,
            'vnf_pkg_id': self.vnf_pkg_id,
            'vnf_metadata': self.vnf_metadata}

        if (self.instantiation_state == fields.VnfInstanceState.INSTANTIATED
                and self.instantiated_vnf_info):
            data.update({'instantiated_vnf_info':
                         self.instantiated_vnf_info.to_dict()})

            vim_connection_info_list = []
            for vim_connection_info in self.vim_connection_info:
                vim_connection_info_list.append(vim_connection_info.to_dict())
            data.update({'vim_connection_info': vim_connection_info_list})

        return data

    @base.remotable
    def update(
            self,
            context,
            vnf_lcm_opoccs,
            body_data,
            vnfd_pkg_data,
            vnfd_id):

        # update vnf_instances
        return _update_vnf_instances(
            context,
            vnf_lcm_opoccs,
            body_data,
            vnfd_pkg_data,
            vnfd_id)

    @base.remotable_classmethod
    def get_by_id(cls, context, id, read_deleted="no"):
        expected_attrs = ["instantiated_vnf_info"]
        db_vnf_instance = _vnf_instance_get_by_id(
            context, id, columns_to_join=expected_attrs,
            read_deleted=read_deleted)
        return cls._from_db_object(context, cls(), db_vnf_instance,
                                   expected_attrs=expected_attrs)


@base.TackerObjectRegistry.register
class VnfInstanceList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfInstance')
    }

    @base.remotable_classmethod
    def get_all(cls, context, expected_attrs=None):
        expected_attrs = ["instantiated_vnf_info"]
        db_vnf_instances = _vnf_instance_list(context,
                                              columns_to_join=expected_attrs)
        return _make_vnf_instance_list(context, cls(), db_vnf_instances,
                                       expected_attrs)

    @base.remotable_classmethod
    def get_by_filters(cls, context, filters=None,
                       expected_attrs=None):
        expected_attrs = ["instantiated_vnf_info"]
        db_vnf_instances = _vnf_instance_list_by_filter(
            context, columns_to_join=expected_attrs,
            filters=filters)

        return _make_vnf_instance_list(context, cls(), db_vnf_instances,
                                       expected_attrs)

    @base.remotable_classmethod
    def vnf_instance_list(cls, vnfd_id, context):
        # get vnf_instance data
        expected_attrs = ["instantiated_vnf_info"]
        db_vnf_instances = _vnf_instance_get(context, vnfd_id,
                                             columns_to_join=expected_attrs)

        vnf_instance_cls = VnfInstance
        vnf_instance_data = ""
        vnf_instance_obj = vnf_instance_cls._from_db_object(
            context, vnf_instance_cls(context), db_vnf_instances,
            expected_attrs=expected_attrs)
        vnf_instance_data = vnf_instance_obj

        return vnf_instance_data
